# -*- coding: utf-8 -*-
"""
dl.py
"""

import torch
from torchvision import transforms
from torch.utils.data import DataLoader

# dataset
# E : 誘電率/導電率分布, 3d配列
# D : レーダー画像, 3d配列
def dataset(ndata, E, D, train_ratio):
    # データ数
    #print(E.shape, D.shape)
    assert E.shape[0] == D.shape[0]
    ndata = D.shape[0] if (ndata <= 0 or ndata > D.shape[0]) else ndata
    #print(ndata)

    # 正規化
    #print(D.mean(), D.std())
    D = (D - D.mean()) / D.std()

    # 乱数配列作成
    rand = torch.rand(ndata)

    # 画像サイズ変更
    #transform_resize = transforms.Resize(size=(56, 56))
    transform_resize = transforms.Resize(size=(112, 112))
    #transform_resize = transforms.Resize(size=(224, 224))
    #transform_flip = transforms.RandomHorizontalFlip(p=1)

    # dataset
    train_set, test_set = [], []
    for i in range(ndata):
        # レーダー画像
        image0 = torch.tensor(D[i])
        image = torch.permute(image0, (1, 0, 2))  # 受信点を最初に持ってくる
        
        # 3D, 受信点削減, NRxに応じて適当に編集する, 全部使用するときは何もしない
        # 以下はNRx=5のとき
        #image = image[0:5:2, :, :]   # 0,2,4  Rx-3
        #image = image[1:4:2, :, :]   # 1,3    Rx-2
        #image = image[2:3, :, :]     # 2      Rx-1

        # resize
        image = transform_resize(image)

        #　誘電率/導電率分布
        label0 = torch.tensor(E[i])
        label = label0.reshape(-1)  # 1D化 x.view(-1)はNG

        # 乱数で分配
        if rand[i] < train_ratio:
            # 訓練データ
            train_set.append([image, label])
        else:
            # テストデータ
            #r = 0.1  # ノイズ
            #image += r * torch.randn(image.shape)
            test_set.append([image, label])

    return train_set, test_set

# dataloader
def dataloader(train_set, test_set, batch_size=50):
    train_loader, test_loader = None, None
    if train_set is not None and len(train_set) > 0:
        train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True)
    if test_set is not None and len(test_set) > 0:
        test_loader  = DataLoader(test_set,  batch_size=batch_size, shuffle=False)

    return train_loader, test_loader
