# -*- coding: utf-8 -*-
"""
plotutils.py
"""

import numpy as np
import matplotlib.pyplot as plt

# 等高線図
# figsize : ウィンドウの幅と高さ(inch)
# mx, my, mstep : 横の数,　縦の数, 関数配列のステップ数
# f[ndata, nx, ny] : 関数配列
# vmin, vmax : 最小, 最大
# sta[ndata] : タイトル配列
# figtitle : ページタイトル
# flipud : 上下反転
def contour(figtitle, figsize, mx, my, mstep, f, vmin, vmax, sta, cmap='rainbow', flipud=False):
    plt.figure(figtitle, figsize=figsize, layout='tight')

    num = 0
    for ix in range(mx):
        for iy in range(my):
            fid = (num + 1) * mstep - 1
            if (fid < f.shape[0]):
                pos = ix + (my - 1 - iy) * mx + 1
                #fid = num * mstep
                #print(Ndata, ix, iy, num, pos)
                ax = plt.subplot(my, mx, pos)
                g = np.transpose(f[fid], (1, 0))
                ax.imshow(g, vmin=vmin, vmax=vmax, cmap=cmap)
                #if flipud:
                #    g = np.flipud(g)
                #ax.pcolormesh(g, vmin=vmin, vmax=vmax, cmap='RdBu_r')#'rainbow')
                #ax.set_aspect('equal')
                #ax.set_axis_off()
                ax.set_xticks([])
                ax.set_yticks([])
                if sta is not None:
                    #title = '%d %.4f' % (fid + 1, sta[fid])
                    ax.set_title(sta[fid], fontsize=8, pad=0)
                num += 1

    suptitle = 'data=%d fig=%d step=%d Nx=%d Ny=%d min=%.5f max=%.5f' % \
        (f.shape[0], num, mstep, f.shape[1], f.shape[2], vmin, vmax)
    #print(suptitle)
    plt.suptitle(suptitle)

    plt.show()

def contour2(figtitle, figsize, mx, my, mstep, f1, f2, vmin, vmax, sta1, sta2):
    assert f1.shape == f2.shape
    plt.figure(figtitle, figsize=figsize, layout='tight')

    num = 0
    for ix in range(mx):
        for iy in range(my):
            fid = (num + 1) * mstep - 1
            if (fid < f1.shape[0]):
                pos1 = (2 * ix + 0) + (my - 1 - iy) * 2 * mx + 1
                ax1 = plt.subplot(my, 2 * mx, pos1)
                #ax1.contourf(np.transpose(f1[fid], (1, 0)), vmin=vmin, vmax=vmax, cmap='rainbow', levels=15)
                ax1.pcolormesh(np.transpose(f1[fid], (1, 0)), vmin=vmin, vmax=vmax, cmap='rainbow')
                ax1.set_aspect('equal')
                ax1.set_axis_off()
                if sta1 is not None:
                    ax1.set_title(sta1[fid], fontsize=8, pad=0, color='r')

                pos2 = (2 * ix + 1) + (my - 1 - iy) * 2 * mx + 1
                ax2 = plt.subplot(my, 2 * mx, pos2)
                #ax2.contourf(np.transpose(f2[fid], (1, 0)), vmin=vmin, vmax=vmax, cmap='rainbow', levels=15)
                ax2.pcolormesh(np.transpose(f2[fid], (1, 0)), vmin=vmin, vmax=vmax, cmap='rainbow')
                ax2.set_aspect('equal')
                ax2.set_axis_off()
                if sta2 is not None:
                    ax2.set_title(sta2[fid], fontsize=8, pad=0)

                num += 1

    suptitle = 'total=%d step=%d fig=%dx2 Nx=%d Ny=%d min=%.5f max=%.5f' % \
        (f1.shape[0], mstep, num, f1.shape[1], f1.shape[2], vmin, vmax)
    plt.suptitle(suptitle)

    plt.show()

# 3D配列を3D表示する, Zスライス, f[i, j, k], 座標軸は描かない
def plot3d_surface(figtitle, f, figsize=(5, 7), stride=1, zfactor=5, cmap='rainbow', elev=15, azim=250):
    assert f.ndim == 3
    ax = plt.figure(figtitle, figsize=figsize, layout='tight').add_subplot(projection='3d')

    cmap = plt.get_cmap(cmap)
    f = f.transpose((2, 0, 1))  # ZXYに変更する
    vmin, vmax = np.min(f), np.max(f)
    nz, nx, ny = f.shape
    x, y = np.meshgrid(np.linspace(0, nx - 1, nx), np.linspace(0, ny - 1, ny))

    for k in range(0, nz, stride):
        z = zfactor * k * np.ones_like(x)
        facecolors = cmap((f[k].transpose(1, 0) - vmin) / (vmax - vmin))
        ax.plot_surface(x, y, z, facecolors=facecolors)

    ax.set_aspect('equal')
    #ax.set_xlabel('X'); ax.set_ylabel('Y'); ax.set_zlabel('Z')
    ax.set_xticks([]); ax.set_yticks([]); ax.set_zticks([])
    ax.view_init(elev=elev, azim=azim, roll=0)

    plt.show()
