/*
GPR
FDTD (2D)
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <time.h>

#ifdef _OPENMP
#include <omp.h>
#endif

#include "alloc.h"

// prototypes
extern void make_mesh(int, int, int, float, float, float *, float *, float *, float *, float *, float *, float *);
extern void tx(int, int, int, int, int, int *, int *);
extern void geometry(int, int, int, int, int, int, int64_t, const int [], const float [], const float [], float, float, const float [], const float [], float, const int [], float ***, float ***, int **);
extern void setupYee(int, int, int, float **, float **, float ***);
extern void setupFactor(int, int, int, float, float ***, float **, float **);
extern void setupMur(int, int, float *, float *, float, float ***, float ***, float ***);
extern void setupPML(int, int, int, float, float *, float *, float *, float *, float *, float *, int **, int **, int **, float *, float *, float *, float *, float *, float *, float *, float *);
extern void dftfactor(int, float, float, int, float *, float *);
extern void solve(int, int, int, float, float, int, int, int, float, int, int, int, int, float **, float **, float **, float *, float *, float *, float *, float ***, float **, float **, float ***, float ***, float **, float **, int **, int **, int **, float *, float *, float *, float *, float *, float *, float *, float *, float *, int, float *, float *, float **, float **);
extern void writedata(int, int, int, int, int, int, float ***, float ***, const char []);
extern void plot2d_g(int, int, int, int, int, int, float ***, int, int, int, int *, int *, const char []);
extern void plot2d_e(int, int, int, int, int, int, int, float ***, float ***, int, int, int *, int *, int *, const char []);
extern void plot2d_r(int, int, int, int, int, int, float ***, int, int, const char []);
extern float average3d(int, int, int, int, int, int, int, float ***);
extern int memory_size(int, int, int, int, int, int, int, int);
extern int file_size(int, int, int, int, int, int, int);
extern void mpi_init(int, char *[], int *, int *);
extern void mpi_close(void);
extern void comm_R(int, int, int, int, int, int, float ***);
extern double cputime(void);

const float C = 2.99792458e8f;  // 光速

int main(int argc, char *argv[])
{
	const int   Ndata   = 1000;           // データ数
	const int   Nrect   = 3;              // 長方形数
	const int   Lrect[] = {5, 15, 10};    // 長方形パラメーター(最小,最大,余白)（セル数）
	const int   Nsmot   = 1;              // 平滑化回数(0以上)
	const int   Nx      = 80;             // X方向セル数
	const int   Ny      = 55;             // Y方向セル数
	const int   iTx0    = 10;             // 送信アンテナ位置の開始点（セル数）
	const int   NTx     = 56;             // 送信アンテナ数
	const int   DTx     = 1;              // 送信アンテナ移動間隔（セル数）
	const int   LTxRx   = 5;              // 送受信アンテナ間距離（セル数）
	const int   Lyant   = 5;              // 送信アンテナの地面からの高さ（セル数）
	const int   Ly0     = 15;             // 地上のセル数
	const float Dx      = 0.025f;         // X方向セルサイズ
	const float Dy      = 0.025f;         // Y方向セルサイズ
	const float Freq    = 600e6f;         // 周波数
	const int   WF      = 0;              // 0:ガウス, 1:微分ガウス
	const int   Nant    = 1;              // 送信アンテナ数（通常1）
	const int   Labc    = 6;              // PML層数（0のときはMur）
	const int   Ntime   = 224;            // タイムステップ数（初期分除く）
	const float Eps0    = 1.0f;           // 地上の比誘電率
	const float Sig0    = 0.0f;           // 地上の導電率[S/m]

	// case-1
	const float Eps[]   = {1.0f, 1.0f};   // 長方形比誘電率(最小,最大)
	const float Sig[]   = {0.0f, 0.0f};   // 長方形導電率[S/m](最小,最大)
	const float pLy1    = 0.0f;           // 地下境界面の存在する確率
	const int   iLy1[]  = {10, 30};       // 地下境界面深さの下限と上限（セル数）
	const float Eps1[]  = {4.0f, 4.0f};   // 地下の比誘電率(最小,最大)
	const float Sig1[]  = {0.0f, 0.0f};   // 地下の導電率[S/m](最小,最大)

/*
	// case-2
	const float Eps[]   = {1.0f, 6.0f};   // 長方形比誘電率(最小,最大)
	const float Sig[]   = {0.0f, 0.03f};  // 長方形導電率[S/m](最小,最大)
	const float pLy1    = 0.0f;           // 地下境界面の存在する確率
	const int   iLy1[]  = {10, 30};       // 地下境界面深さの下限と上限（セル数）
	const float Eps1[]  = {4.0f, 4.0f};   // 地下の比誘電率(最小,最大)
	const float Sig1[]  = {0.0f, 0.0f};   // 地下の導電率[S/m](最小,最大)
*/
/*
	// case-3
	const float Eps[]   = {1.0f, 3.0f};   // 長方形比誘電率(最小,最大)
	const float Sig[]   = {0.0f, 0.01f};  // 長方形導電率[S/m](最小,最大)
	const float pLy1    = 0.3f;           // 地下境界面の存在する確率
	const int   iLy1[]  = {10, 30};       // 地下境界面深さの下限と上限（セル数）
	const float Eps1[]  = {3.0f, 4.0f};   // 地下の比誘電率(最小,最大)
	const float Sig1[]  = {0.0f, 0.01f};  // 地下の導電率[S/m](最小,最大)
*/
/*
	// case-4
	const float Eps[]   = {1.0f, 1.0f};   // 長方形比誘電率(最小,最大)
	const float Sig[]   = {0.0f, 1.0f};   // 長方形導電率[S/m](最小,最大)
	const float pLy1    = 0.0f;           // 地下境界面の存在する確率
	const int   iLy1[]  = {10, 30};       // 地下境界面深さの下限と上限（セル数）
	const float Eps1[]  = {4.0f, 4.0f};   // 地下の比誘電率(最小,最大)
	const float Sig1[]  = {0.0f, 0.0f};   // 地下の導電率[S/m](最小,最大)
*/
	const int Ly = Ny - Ly0;              // 地下のセル数
	const int64_t Seed = 10000;           // 乱数の種, 10000:学習用, 20000:テスト用
	char outfile[BUFSIZ] = "fdtd.bin";    // 出力ファイル名

	// debug
	const int log1 = 1;  // (数値出力) 計算条件(0/1, 通常1)
	const int log2 = 0;  // (数値出力) 収束状況(0/1, 通常0)
	const int fig1 = 0;  // (図形出力) 誘電率/導電率分布(0/1, 通常0)
	const int fig2 = 0;  // (図形出力) 電界分布(0/1, 通常0, 計算時間が増えるので必要なときのみ)
	const int fig3 = 0;  // (図形出力) レーダー画像(0/1, 通常0)
	const char fig1ev2[] = "fig1.ev2";  // $ ev2d.exe fig1.ev2
	const char fig2ev2[] = "fig2.ev2";  // $ ev2d.exe fig2.ev2
	const char fig3ev2[] = "fig3.ev2";  // $ ev2d.exe fig3.ev2

	// MPI
	int commsize, commrank;
	mpi_init(argc, argv, &commsize, &commrank);
	//printf("%d %d\n", commsize, commrank); fflush(stdout);

	// 引数: スレッド数、出力ファイル
	int nthread = 1;
	if (argc > 1) {
		nthread = atoi(argv[1]);
	}
	if (argc > 2) {
		strcpy(outfile, argv[2]);
	}

	// OpenMP
#ifdef _OPENMP
	omp_set_num_threads(nthread);
#endif

	// 開始時刻
	const double tstart = cputime();

	// メッシュ作成
	float Dt;
	float *Xn, *Yn, *dXn, *dYn, *dXc, *dYc;
	alloc1d(float, Xn,  Nx + 1)
	alloc1d(float, Yn,  Ny + 1)
	alloc1d(float, dXn, Nx + 1)
	alloc1d(float, dYn, Ny + 1)
	alloc1d(float, dXc, Nx + 0)
	alloc1d(float, dYc, Ny + 0)
	make_mesh(Nx, Ny, Ly, Dx, Dy, Xn, Yn, &Dt, dXn, dYn, dXc, dYc);

	// 送信点座標
	int *iTx, *jTx;
	alloc1d(int, iTx, NTx)
	alloc1d(int, jTx, NTx)
	tx(NTx, iTx0, DTx, Ly, Lyant, iTx, jTx);

	// タイムステップ数 (送信点によらない)
	const float t0 = (float)(2.0 * (1.27 / Freq) + (2 * Lyant * Dy * sqrt(Eps0)) / C);  // 波形記録開始時刻
	const int it0 = (int)(t0 / Dt);
	const int Niter = Ntime + it0;
	//printf("%e %e %d %d %d\n", t0, Dt, it0, Ntime, Niter);

	// 計算条件表示
	if (log1 && (commrank == 0)) {
		const int mem = memory_size(nthread, Ndata, Nx, Ny, NTx, Niter, Labc, fig2);
		const int filesize = file_size(Ndata, Nx, Ly, 1, NTx, 1, Ntime);
		printf("FDTD(2D) data=%d Nx=%d Ny=%d Tx=%d rect=%d(%d) WF=%d ABC=%d dt=%.5fnsec\n",
			Ndata, Nx, Ny, NTx, Nrect, Nsmot, WF, Labc, Dt * 1e9);
		printf("process=%d thread=%d memory=%dMB iter=%d(%d) output=%s(%dMB)\n",
			commsize, nthread, mem, Niter, Ntime, outfile, filesize);
		fflush(stdout);
	}

	// 長方形(debug)
	int **rect;
	alloc2d(int, rect, Ndata, 4 * Nrect)

	// 誘電率配列(セル中心, xデータ数)
	float ***Epsr, ***Sigm;
	alloc3d(float, Epsr, Ndata, Nx, Ny)
	alloc3d(float, Sigm, Ndata, Nx, Ny)

	// 誘電率配列代入
	//geometry(Ndata, Nrect, Seed, Lrect, Eps, Sig, Nsmot, Nx, Ly1, Ly2, Ly3, Epsr, Sigm, Eps1, Sig1, Eps2, Sig2, Eps3, Sig3, rect);
	geometry(Ndata, Nx, Ny, Ly, Nrect, Nsmot, Seed,
		Lrect, Eps, Sig,
		Eps0, Sig0, Eps1, Sig1, pLy1, iLy1,
		Epsr, Sigm, rect);

	// debug, 全領域誘電率/導電率分布図
	if (((fig1 == 1) || (fig1 == -1)) && (commrank == 0)) {
		plot2d_g(1, 1, Ndata, 1, Nx, Ny, Epsr, 1, 0, NTx, iTx, jTx, fig1ev2);  // 誘電率
		//plot2d_g(1, 1, Ndata, 1, Nx, Ny, Sigm, 1, 0, NTx, iTx, jTx, fig1ev2);  // 導電率
	}
	if (fig1 < 0) {
		exit(0);  // debug: 計算せずに終了する
	}

	// レーダー画像（データ数x送信点数xタイムステップ数）
	float ***Er = NULL;
	alloc3d(float, Er, Ndata, NTx, Niter)

	// debug, 全領域調和界（PML層を含まない, x送信点, データ0のみ）
	float ***Ez_r = NULL, ***Ez_i = NULL;
	if (fig2 && (commrank == 0)) {
		alloc3d(float, Ez_r, NTx, Nx + 1, Ny + 1)
		alloc3d(float, Ez_i, NTx, Nx + 1, Ny + 1)
	}

	// PML用配列（データに無関係, readonly）
	int **bEzpml = NULL, **bHxpml = NULL, **bHypml = NULL;
	float *fEzpmlx = NULL, *fEzpmly = NULL, *fHxpml = NULL, *fHypml = NULL;
	float *dXnpml = NULL, *dXcpml = NULL, *dYnpml = NULL, *dYcpml = NULL;
	if (Labc > 0) {
		alloc2d(int, bEzpml, Nx + 0 + 2 * Labc, Ny + 0 + 2 * Labc)
		alloc2d(int, bHxpml, Nx + 0 + 2 * Labc, Ny + 0 + 2 * Labc)
		alloc2d(int, bHypml, Nx + 0 + 2 * Labc, Ny + 0 + 2 * Labc)
		alloc1d(float, fEzpmlx, Nx + 1 + 2 * Labc)
		alloc1d(float, fEzpmly, Ny + 1 + 2 * Labc)
		alloc1d(float, fHxpml,  Ny + 0 + 2 * Labc)
		alloc1d(float, fHypml,  Nx + 0 + 2 * Labc)
		alloc1d(float, dXnpml,  Nx + 0 + 2 * Labc)
		alloc1d(float, dXcpml,  Nx + 0 + 2 * Labc)
		alloc1d(float, dYnpml,  Ny + 0 + 2 * Labc)
		alloc1d(float, dYcpml,  Ny + 0 + 2 * Labc)
		setupPML(Nx, Ny, Labc, Dt, Xn, Yn, dXn, dYn, dXc, dYc, bEzpml, bHxpml, bHypml, fEzpmlx, fEzpmly, fHxpml, fHypml, dXnpml, dXcpml, dYnpml, dYcpml);
	}

	// DFT係数, debug
	float *Fcos = NULL, *Fsin = NULL;
	if (fig2 && (commrank == 0)) {
		alloc1d(float, Fcos, Niter + 1)
		alloc1d(float, Fsin, Niter + 1)
		dftfactor(Niter, Freq, Dt, WF, Fcos, Fsin);
	}

	// プロセスごとのデータ番号の範囲（ブロック分割）
	const int ndata = (Ndata + (commsize - 1)) / commsize;
	int idata0 = commrank * ndata;
	if (idata0 > Ndata - 1) {
		idata0 = Ndata - 1;
	}
	int idata1 = idata0 + ndata;
	if (idata1 > Ndata) {
		idata1 = Ndata;
	}
	//printf("%d %d %d %d %d %d\n", commsize, commrank, Ndata, ndata, idata0, idata1); fflush(stdout);

	// OpenMP並列領域
#ifdef _OPENMP
#pragma omp parallel
#endif
{
	// Yee格子物性値(スレッド変数, 磁性体は考えないので電界成分のみ)
	float ***PEz = NULL;
	alloc3d(float, PEz, Nx + 1 + 2 * Labc, Ny + 1 + 2 * Labc, 2)

	// FDTD係数(スレッド変数, 磁性体は考えないので電界成分のみ)
	float **C1Ez = NULL, **C2Ez = NULL;
	alloc2d(float, C1Ez, Nx + 1, Ny + 1)
	alloc2d(float, C2Ez, Nx + 1, Ny + 1)

	// ABC変数(スレッド変数)
	float ***Ezmurx = NULL, ***Ezmury = NULL;
	float **Ezpmlx = NULL, **Ezpmly = NULL;
	if (Labc == 0) {
		alloc3d(float, Ezmurx, Nx + 1, Ny + 1, 2)
		alloc3d(float, Ezmury, Nx + 1, Ny + 1, 2)
	}
	else {
		alloc2d(float, Ezpmlx, Nx + 1 + 2 * Labc, Ny + 1 + 2 * Labc)
		alloc2d(float, Ezpmly, Nx + 1 + 2 * Labc, Ny + 1 + 2 * Labc)
	}

	// 電磁界配列(スレッド変数)
	float **Ez = NULL, **Hx = NULL, **Hy = NULL;
	alloc2d(float, Ez, Nx + 1 + 2 * Labc, Ny + 1 + 2 * Labc)
	alloc2d(float, Hx, Nx + 1 + 2 * Labc, Ny + 0 + 2 * Labc)
	alloc2d(float, Hy, Nx + 0 + 2 * Labc, Ny + 1 + 2 * Labc)

	// スレッド番号
	int ithread = 0;
#ifdef _OPENMP
	ithread = omp_get_thread_num();
#endif

	// データに関するループ
	//for (int idata = 0; idata < Ndata; idata++) {
	for (int idata = idata0 + ithread; idata < idata1; idata += nthread) {
		//printf("%d %d %d\n", commrank, ithread, idata); fflush(stdout);
		assert((idata >= 0) && (idata < Ndata));
		// 出力用フラグ
		const int ilog = log2 && (commrank == 0);  // 収束状況
		const int iout = fig2 && (commrank == 0) && (idata == 0);  // 電界分布（データ0のみ）
		if (log1 && (idata % 1000 == 0)) {printf("%d %d\n", idata, ithread); fflush(stdout);}  // 経過出力

		// Yee格子物性値（データに依存）
		setupYee(Nx, Ny, Labc, Epsr[idata], Sigm[idata], PEz);

		// FDTD係数（データに依存）
		setupFactor(Nx, Ny, Labc, Dt, PEz, C1Ez, C2Ez);

		// Mur係数（データに依存）
		if (Labc == 0) {
			setupMur(Nx, Ny, Xn, Yn, Dt, PEz, Ezmurx, Ezmury);
		}

		// 送信点に関するループ
		for (int itx = 0; itx < NTx; itx++) {
			solve(
				Niter, Nx, Ny, Freq, Dt, Labc, WF, Nant, Dy, ilog,
				iTx[itx], jTx[itx], LTxRx,
				Ez, Hx, Hy,
				dXn, dYn, dXc, dYc,
				PEz, C1Ez, C2Ez,
				Ezmurx, Ezmury,
				Ezpmlx, Ezpmly, bEzpml, bHxpml, bHypml,
				fEzpmlx, fEzpmly, fHxpml, fHypml,
				dXnpml, dXcpml, dYnpml, dYcpml,
				Er[idata][itx],
				iout, Fcos, Fsin, (iout ? Ez_r[itx] : NULL), (iout ? Ez_i[itx] : NULL));
		}
	}
}
	// レーダ画像をrootに集める
	if (commsize > 1) {
		comm_R(commsize, commrank, idata0, idata1, NTx, Niter, Er);
	}

	// 出力:root
	if (commrank == 0) {
		// ファイル出力, ファイルサイズが大きいので一方のみを選ぶ
		writedata(Ndata, Nx, Ly, NTx, Niter, it0, Epsr, Er, outfile);  // 誘電率
		//writedata(Ndata, Nx, Ly, NTx, Niter, it0, Sigm, Er, outfile);  // 導電率

		// debug, レーダー画像平均数値出力
		if (log1) {
			const float rave = average3d(0, Ndata, 0, NTx, 0, Niter, 1, Er);
			printf("Rave=%.8e\n", rave);
		}

		// debug, 電界分布図形出力、データ0
		if (fig2) {
			plot2d_e(1, 1, NTx, 1, Nx, Ny, Ly, Ez_r, Ez_i, 50, Nrect, rect[0], iTx, jTx, fig2ev2);
		}

		// debug, レーダー画像図形出力
		if (fig3) {
			plot2d_r(1, 1, Ndata, 1, NTx, Niter, Er, it0, 2, fig3ev2);
		}
	}

	// 終了時刻
	const double tstop = cputime();
	if (log1 && (commrank == 0)) {
		printf("CPU=%.3f[sec]\n", tstop - tstart);
		fflush(stdout);
	}

	// MPI
	mpi_close();

	return 0;
}
