/*
mur.c
Mur-ABC
磁性体は考えない
*/

#include <math.h>

extern float C;

// Ez係数
void setupMur(int Nx, int Ny, float *Xn, float *Yn, float Dt, float ***PEz, float ***Ezmurx, float ***Ezmury)
{
	// X境界
	for (int i = 0; i < Nx + 1; i += Nx) {
		const float d = (i == 0) ? (Xn[1] - Xn[0]) : (Xn[Nx] - Xn[Nx - 1]);
		for (int j = 0; j < Ny + 1; j++) {
			const float v = C * (float)sqrt(PEz[i][j][0]);
			Ezmurx[i][j][1] = (v * Dt - d) / (v * Dt + d);
		}
	}

	// Y境界
	for (int j = 0; j < Ny + 1; j += Ny) {
		const float d = (j == 0) ? (Yn[1] - Yn[0]) : (Yn[Ny] - Yn[Ny - 1]);
		for (int i = 0; i < Nx + 1; i++) {
			const float v = C * (float)sqrt(PEz[i][j][0]);
			Ezmury[i][j][1] = (v * Dt - d) / (v * Dt + d);
		}
	}
}


// Ez更新
void murEz(int Nx, int Ny, float **Ez, float ***Ezmurx, float ***Ezmury)
{
	// X境界
	for (int j = 0; j < Ny + 1; j++) {
		Ez[ 0][j] = Ezmurx[ 0][j][0] + Ezmurx[ 0][j][1] * (Ez[     1][j] - Ez[ 0][j]);
		Ez[Nx][j] = Ezmurx[Nx][j][0] + Ezmurx[Nx][j][1] * (Ez[Nx - 1][j] - Ez[Nx][j]);
		Ezmurx[ 0][j][0] = Ez[     1][j];
		Ezmurx[Nx][j][0] = Ez[Nx - 1][j];
	}

	// Y境界
	for (int i = 0; i < Nx + 1; i++) {
		Ez[i][ 0] = Ezmury[i][ 0][0] + Ezmury[i][ 0][1] * (Ez[i][     1] - Ez[i][ 0]);
		Ez[i][Ny] = Ezmury[i][Ny][0] + Ezmury[i][Ny][1] * (Ez[i][Ny - 1] - Ez[i][Ny]);
		Ezmury[i][ 0][0] = Ez[i][     1];
		Ezmury[i][Ny][0] = Ez[i][Ny - 1];
	}
}
