/*
plot2d.c
fdtd2d
2D図形出力, debug用
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "ev.h"
#include "alloc.h"

extern void getminmax2d(int, int, int, int, int, float **, float *, float *);
extern void getminmax3d(int, int, int, int, int, int, int, float ***, float *, float *);
extern void plot2d_contour(float, float, int, int, int, int, float **, float, float, int, int);

// 誘電率分布
void plot2d_g(int mx, int my, int ndata, int nstep, int nx, int ny,
	float ***d, int color, int grid, int ntx, int *itx, int *jtx, const char fn[])
{
	char str[BUFSIZ], fmt[BUFSIZ];
	const float dm = 0.4f;

	// 最小・最大
	float dmin, dmax;
	getminmax3d(0, ndata, 0, nx, 0, ny, 0, d, &dmin, &dmax);

	// 開始
	const float s = 4;
	const float w = (mx * nx) + ((mx + 1) * s);
	const float h = (my * ny) + ((my + 1) * s);
	ev2d_init(w, h);
	ev2d_newPage();

	// 等高線図, 1データ=1図
	int num = 0;
	for (int px = 0; px < mx; px++) {
	for (int py = 0; py < my; py++) {
		if (num * nstep < ndata) {
			// 原点
			const float x0 = (px * nx) + ((px + 1) * s);
			const float y0 = (py * ny) + ((py + 1) * s);

			// 等高線図
			plot2d_contour(x0, y0, 0, nx, 0, ny, d[num * nstep], dmin, dmax, color, grid);

			// データ番号
			ev2d_setColor(0, 0, 0);
			sprintf(str, "%d", num * nstep + 1);
			ev2d_drawString(x0, y0 + ny - h / 60, h / 60, str);

			// 送信点
			ev2d_setColor(0, 0, 0);
			for (int n = 0; n < ntx; n++) {
				ev2d_drawRectangle(x0 + itx[n] - dm, y0 + jtx[n] - dm, x0 + itx[n] + dm, y0 + jtx[n] + dm);
			}

			// 図++
			num++;
		}
	}
	}

	// 最小・最大
	strcpy(fmt, "data=%d fig=%d step=%d Nx=%d Ny=%d Tx=%d min=%.4f max=%.4f");
	sprintf(str, fmt, ndata, num, nstep, nx, ny, ntx, dmin, dmax);
	ev2d_setColor(0, 0, 0);
	ev2d_drawString(s, s, h / 30, str);

	// 出力
	ev2d_file(1, fn);
	ev2d_output();
}


// 電界分布
void plot2d_e(int mx, int my, int ntx, int nstep, int nx, int ny, int ly,
	float ***e_r, float ***e_i, int db, int nrect, int *rect, int *iTx, int *jTx, const char fn[])
{
	char str[BUFSIZ], fmt[BUFSIZ];
	const float eps = 1e-12f;
	const float dm = 0.5f;
	//printf("%d %d %d %d %d %d %d\n", mx, my, ntx, nstep, nx, ny, ly);
	assert((mx > 0) && (my > 0) && (ntx > 0) && (nstep > 0) && (nx > 0) && (ny > 0) && (ly > 0) && (nrect >= 0));

	// 節点->セル中心, 複素数->絶対値
	float ***d;
	alloc3d(float, d, ntx, nx, ny)
	for (int itx = 0; itx < ntx; itx++) {
	for (int i = 0; i < nx; i++) {
	for (int j = 0; j < ny; j++) {
		const float d_r = (e_r[itx][i][j] + e_r[itx][i + 1][j] + e_r[itx][i][j + 1] + e_r[itx][i + 1][j + 1]) / 4;
		const float d_i = (e_i[itx][i][j] + e_i[itx][i + 1][j] + e_i[itx][i][j + 1] + e_i[itx][i + 1][j + 1]) / 4;
		d[itx][i][j] = (float)sqrt((d_r * d_r) + (d_i * d_i));
	}
	}
	}

	// 最小・最大
	float dmin, dmax;
	getminmax3d(0, ntx, 0, nx, 0, ny, db, d, &dmin, &dmax);
	//printf("%f %f\n", dmin, dmax);
	if (!db) {
		dmin = 0;
	}

	// dB
	if (db) {
		for (int itx = 0; itx < ntx; itx++) {
		for (int i = 0; i < nx; i++) {
		for (int j = 0; j < ny; j++) {
			d[itx][i][j] = 20 * (float)log10(fmax(d[itx][i][j], eps));
		}
		}
		}
	}

	// 開始
	const float s = 2;
	const float w = (mx * nx) + ((mx + 1) * s);
	const float h = (my * ny) + ((my + 1) * s);
	ev2d_init(w, h);
	ev2d_newPage();

	// 等高線図, 1送信点=1図
	int itx = nstep - 1;
	for (int px = 0; px < mx; px++) {
	for (int py = 0; py < my; py++) {
		if (itx < ntx) {
			// 原点
			const float x0 = (px * nx) + ((px + 1) * s);
			const float y0 = (py * ny) + ((py + 1) * s);
			//printf("%d %d %d %d %f %f\n", mx, my, px, py, x0, y0);

			// 等高線図(カラー)
			plot2d_contour(x0, y0, 0, nx, 0, ny, d[itx], dmin, dmax, 1, 0);

			// 長方形
			ev2d_setColor(0, 0, 0);
			for (int irect = 0; irect < nrect; irect++) {
				int *ptr = &rect[4 * irect];
				ev2d_drawRectangle(x0 + ptr[0], y0 + ptr[1], x0 + ptr[2], y0 + ptr[3]);
			}

			// 境界面
			ev2d_setColor(0, 0, 0);
			ev2d_drawLine(x0, y0 + ly, x0 + nx, y0 + ly);

			// 送信点番号
			ev2d_setColor(0, 0, 0);
			sprintf(str, "%d", itx + 1);
			ev2d_drawString(x0, y0 + ny - h / 60, h / 60, str);

			// 送信点
			ev2d_setColor(0, 0, 0);
			ev2d_fillRectangle(x0 + iTx[itx] - dm, y0 + jTx[itx] - dm, x0 + iTx[itx] + dm, y0 + jTx[itx] + dm);

			// 送信点++
			itx += nstep;
		}
	}
	}

	// 最小・最大
	strcpy(fmt, db ? "Tx=%d Nx=%d Ny=%d min=%.3fdB max=%.3fdB"
	               : "Tx=%d Nx=%d Ny=%d min=%.6f max=%.6f");
	sprintf(str, fmt, ntx, nx, ny, dmin, dmax);
	ev2d_setColor(0, 0, 0);
	ev2d_drawString(s, s, h / 40, str);

	// 出力
	ev2d_file(1, fn);
	ev2d_output();
}


// レーダー画像, 線形
void plot2d_r(int mx, int my, int ndata, int nstep, int ntx, int niter, float ***r, int it0, int icolor, const char fn[])
{
	char str[BUFSIZ], fmt[BUFSIZ];
	assert((mx > 0) && (my > 0) && (ndata > 0) && (nstep > 0) && (ntx > 0) && (niter > 0) && (it0 > 0));
	//printf("%d %d %d %d %d %d\n", mx, my, ndata, nstep, ntx, niter);
	//it0 = 0;

	// 最小・最大
	float dmin, dmax;
	getminmax3d(0, ndata, 0, ntx, it0, niter, 0, r, &dmin, &dmax);
	//printf("%d %f %f\n", wf, dmin, dmax);

	// 最小値=-最大値, 青赤2色に適している
	if (dmin + dmax > 0) {
		dmin = -dmax;
	}
	else {
		dmax = -dmin;
	}

	// 図の大きさ
	const int nx = ntx;
	const int ny = niter - it0;

	// 2D配列
	float **d;
	alloc2d(float, d, nx, ny)

	// 開始
	const float s = 2;
	const float w = (mx * nx) + ((mx + 1) * s);
	const float h = (my * ny) + ((my + 1) * s);
	ev2d_init(w, h);
	ev2d_newPage();

	int idata = nstep - 1;
	for (int px = 0; px < mx; px++) {
	for (int py = 0; py < my; py++) {
		if (idata < ndata) {
			// 関数値, 上下反転
			for (int i = 0; i < nx; i++) {
			for (int j = 0; j < ny; j++) {
				d[i][j] = r[idata][i][niter - 1 - j];
			}
			}

			// 原点
			const float x0 = (px * nx) + ((px + 1) * s);
			const float y0 = (py * ny) + ((py + 1) * s);
			//printf("%d %d %d %d %f %f\n", mx, my, px, py, x0, y0);

			// 等高線図
			plot2d_contour(x0, y0, 0, nx, 0, ny, d, dmin, dmax, icolor, 0);

			// 領域
			ev2d_setColor(0, 0, 0);
			ev2d_drawRectangle(x0, y0, x0 + nx, y0 + ny);

			// データ番号
			ev2d_setColor(0, 0, 0);
			sprintf(str, "%d", idata + 1);
			ev2d_drawString(x0, y0 + ny - h / 60, h / 60, str);

			// データ番号++
			idata += nstep;
		}
	}
	}

	// 最小・最大
	strcpy(fmt, "data=%d step=%d Tx=%d time=%d-%d min=%.6f max=%.6f");
	sprintf(str, fmt, ndata, nstep, ntx, it0, niter, dmin, dmax);
	ev2d_drawString(s, 2 * s, h / 40, str);

	// 出力
	ev2d_file(1, fn);
	ev2d_output();
}
