/*
plotutils.c
fdtd2d/fdtd3d共通
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "ev.h"
#include "alloc.h"

static void signed_color(double, double, double, unsigned char *, unsigned char *, unsigned char *);

// 等高線図
// icolor = 0:モノクロ, 1:カラー, 2:青赤2色
void plot2d_contour(float x0, float y0, int i0, int i1, int j0, int j1,
	float **d, float dmin, float dmax, int icolor, int grid)
{
	//printf("%d %f %f\n", icolor, dmin, dmax);
	// 等高線図（塗りつぶし）
	for (int i = i0; i < i1; i++) {
	for (int j = j0; j < j1; j++) {
		// 色
		if ((icolor == 0) || (icolor == 1)) {
			const float v = (d[i][j] - dmin) / (dmax - dmin);
			ev2d_setColorV(v, icolor);
		}
		else {
			unsigned char r, g, b;
			signed_color(d[i][j], dmin, dmax, &r, &g, &b);
			ev2d_setColor(r, g, b);
		}

		// 長方形塗りつぶし
		ev2d_fillRectangle(x0 + i - i0, y0 + j - j0, x0 + i - i0 + 1, y0 + j - j0 + 1);
	}
	}

	// 外枠（モノクロのとき）
	if (icolor == 0) {
		ev2d_setColor(0, 0, 0);
		ev2d_drawRectangle(x0, y0, x0 + i1 - i0, y0 + j1 - j0);
	}

	// グリッド（内部）
	if (grid) {
		ev2d_setColor(160, 160, 160);
		for (int i = i0 + 1; i < i1; i++) {
			ev2d_drawLine(x0 + i - i0, y0, x0 + i - i0, y0 + j1 - j0);
		}
		for (int j = j0 + 1; j < j1; j++) {
			ev2d_drawLine(x0, y0 + j - j0, x0 + i1 - i0, y0 + j - j0);
		}
	}
}


// 符号付の値の色
static void signed_color(double d, double dmin, double dmax, unsigned char *r, unsigned char *g, unsigned char *b)
{
	const double d0 = (fabs(dmax) > fabs(dmin)) ? fabs(dmax) : fabs(dmin);
	d = d / d0;
	if (d > +1) d = +1;
	if (d < -1) d = -1;
	const double r0 = (d > 0) ? 1     : 1 + d;
	const double g0 = (d > 0) ? 1 - d : 1 + d;
	const double b0 = (d > 0) ? 1 - d : 1;
	*r = (unsigned char)(255 * r0);
	*g = (unsigned char)(255 * g0);
	*b = (unsigned char)(255 * b0);
}
