/*
pml.c
PML-ABC
磁性体は考えない
*/

#include <stdlib.h>
#include <stdio.h>

extern void pmlfactor(int, float *);

extern float C;

// PML係数
void setupPML(int Nx, int Ny, int Labc, float Dt,
	float *Xn, float *Yn, float *dXn, float *dYn, float *dXc, float *dYc,
	int **bEzpml, int **bHxpml, int **bHypml,
	float *fezx, float *fezy, float *fhx, float *fhy,
	float *dxnpml, float *dxcpml, float *dynpml, float *dycpml)
{
	// PML因子
	float *f = (float *)malloc((2 * Labc + 1) * sizeof(float));
	pmlfactor(Labc, f);

	const float cdt = C * Dt;
	float fc = 0;

	// Ez
	for (int i = -Labc; i < Nx + Labc + 1; i++) {
		if      (i <      1) fc = f[- 2 * (i     )    ] / (Xn[1 ] - Xn[0     ]);
		else if (i > Nx - 1) fc = f[+ 2 * (i - Nx)    ] / (Xn[Nx] - Xn[Nx - 1]);
		else                 fc = 0;
		fezx[i + Labc] = 1 / (1 + cdt * fc);
		//printf("%d %e\n", i, fezx[i + Labc]);
	}
	for (int j = -Labc; j < Ny + Labc + 1; j++) {
		if      (j <      1) fc = f[- 2 * (j     )    ] / (Yn[1 ] - Yn[0     ]);
		else if (j > Ny - 1) fc = f[+ 2 * (j - Ny)    ] / (Yn[Ny] - Yn[Ny - 1]);
		else                 fc = 0;
		fezy[j + Labc] = 1 / (1 + cdt * fc);
		//printf("%d %e\n", j, fezy[j + Labc]);
	}

	// Hx
	for (int j = -Labc; j < Ny + Labc + 0; j++) {
		if      (j <      0) fc = f[- 2 * (j     ) - 1] / (Yn[1 ] - Yn[0     ]);
		else if (j > Ny - 1) fc = f[+ 2 * (j - Ny) + 1] / (Yn[Ny] - Yn[Ny - 1]);
		else                 fc = 0;
		fhx[j + Labc] = 1 / (1 + cdt * fc);
		//printf("%d %e\n", j, fhx[j + Labc]);
	}

	// Hy
	for (int i = -Labc; i < Nx + Labc + 0; i++) {
		if      (i <      0) fc = f[- 2 * (i     ) - 1] / (Xn[1 ] - Xn[0     ]);
		else if (i > Nx - 1) fc = f[+ 2 * (i - Nx) + 1] / (Xn[Nx] - Xn[Nx - 1]);
		else                 fc = 0;
		fhy[i + Labc] = 1 / (1 + cdt * fc);
		//printf("%d %e\n", i, fhy[i + Labc]);
	}

	// dx
	for (int i = -Labc + 1; i < Nx + Labc + 0; i++) {
		dxnpml[i + Labc] = dXn[(i < 0) ? 0 : (i > Nx - 0) ? Nx - 0 : i];
	}
	for (int i = -Labc + 0; i < Nx + Labc + 0; i++) {
		dxcpml[i + Labc] = dXc[(i < 0) ? 0 : (i > Nx - 1) ? Nx - 1 : i];
	}
	// dy
	for (int j = -Labc + 1; j < Ny + Labc + 0; j++) {
		dynpml[j + Labc] = dYn[(j < 0) ? 0 : (j > Ny - 0) ? Ny - 0 : j];
	}
	for (int j = -Labc + 0; j < Ny + Labc + 0; j++) {
		dycpml[j + Labc] = dYc[(j < 0) ? 0 : (j > Ny - 1) ? Ny - 1 : j];
	}

	// PML領域内であるか(0/1)
	// Ez
	for (int i = -Labc + 1; i < Nx + Labc + 0; i++) {
	for (int j = -Labc + 1; j < Ny + Labc + 0; j++) {
		bEzpml[i + Labc][j + Labc] = (i < 0) || (Nx - 0 < i) ||
		                             (j < 0) || (Ny - 0 < j);
	}
	}
	// Hx
	for (int i = -Labc + 1; i < Nx + Labc + 0; i++) {
	for (int j = -Labc + 0; j < Ny + Labc + 0; j++) {
		bHxpml[i + Labc][j + Labc] = (i < 0) || (Nx - 0 < i) ||
		                             (j < 0) || (Ny - 1 < j);
	}
	}
	// Hy
	for (int i = -Labc + 0; i < Nx + Labc + 0; i++) {
	for (int j = -Labc + 1; j < Ny + Labc + 0; j++) {
		bHypml[i + Labc][j + Labc] = (i < 0) || (Nx - 1 < i) ||
		                             (j < 0) || (Ny - 0 < j);
	}
	}

	// free
	free(f);
}


// Ez更新
void pmlEz(int Nx, int Ny, int Labc,
	float **Ez, float **Hx, float **Hy, float **Ezx, float **Ezy,
	int **bEzpml, float ***PEz, float *fEzx, float *fEzy, float *dXnpml, float *dYnpml)
{
	const int nx = Nx + 2 * Labc + 0;
	const int ny = Ny + 2 * Labc + 0;

	for (int i = 1; i < nx; i++) {
	for (int j = 1; j < ny; j++) {
		if (bEzpml[i][j]) {
			const float dx = dXnpml[i] * PEz[i][j][0]; 
			const float dhy = Hy[i    ][j    ]
			                - Hy[i - 1][j    ];
			Ezx[i][j] = (
			Ezx[i][j] + dx * dhy) * fEzx[i];

			const float dy = dYnpml[j] * PEz[i][j][0];
			const float dhx = Hx[i    ][j    ]
			                - Hx[i    ][j - 1];
			Ezy[i][j] = (
			Ezy[i][j] - dy * dhx) * fEzy[j];

			Ez[i][j] = Ezx[i][j] + Ezy[i][j];
		}
	}
	}
}


// Hx更新
void pmlHx(int Nx, int Ny, int Labc,
	float **Ez, float **Hx,
	int **bHxpml, float *fHx, float *dYcpml)
{
	const float ramur = 1;  // 磁性体なし
	const int nx = Nx + 2 * Labc + 0;
	const int ny = Ny + 2 * Labc + 0;

	for (int i = 1; i < nx; i++) {
	for (int j = 0; j < ny; j++) {
		if (bHxpml[i][j]) {
			const float dy = dYcpml[j] * ramur;
			const float dez = Ez[i    ][j + 1]
			                - Ez[i    ][j    ];
			Hx[i][j] = (
			Hx[i][j] - dy * dez) * fHx[j];
		}
	}
	}
}


// Hy更新
void pmlHy(int Nx, int Ny, int Labc,
	float **Ez, float **Hy,
	int **bHypml, float *fHy, float *dXcpml)
{
	const float ramur = 1;  // 磁性体なし
	const int nx = Nx + 2 * Labc + 0;
	const int ny = Ny + 2 * Labc + 0;

	for (int i = 0; i < nx; i++) {
	for (int j = 1; j < ny; j++) {
		if (bHypml[i][j]) {
			const float dx = dXcpml[i] * ramur;
			const float dez = Ez[i + 1][j    ]
			                - Ez[i    ][j    ];
			Hy[i][j] = (
			Hy[i][j] + dx * dez) * fHy[i];
		}
	}
	}
}
