/*
writedata.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <assert.h>

extern void getminmax3d(int, int, int, int, int, int, int, float ***, float *, float *);

// ファイル出力
void writedata(int ndata, int nx, int ly, int ntx, int niter, int it0, float ***e, float ***d, const char outfile[])
{
	//printf("%d %d %d %d %d %d\n", ndata, nx, ly, ntx, niter, it0);

	// (1) 誘電率分布(地上除く)

	// 最小値・最大値
	float emin, emax;
	getminmax3d(0, ndata, 0, nx, 0, ly, 0, e, &emin, &emax);
	//printf("%f %f\n", emin, emax);

	// 1D配列化+1バイト化
	const int64_t ne = ndata * nx * ly;
	unsigned char *e1d = (unsigned char *)malloc(ne * sizeof(unsigned char));
	int64_t ie = 0;
	for (int n = 0; n < ndata; n++) {
	for (int i = 0; i < nx; i++) {
	for (int j = 0; j < ly; j++) {
		e1d[ie++] = (unsigned char)(255 * (e[n][i][j] - emin) / (emax - emin));
	}
	}
	}
	assert(ie == ne);

	// (2) レーダー画像(初期除く)

	// 最小値・最大値
	float dmin, dmax;
	getminmax3d(0, ndata, 0, ntx, it0, niter, 0, d, &dmin, &dmax);
	//printf("%f %f\n", dmin, dmax);

	// 1D配列化
	const int64_t nd = ndata * ntx * (niter - it0);
	unsigned char *d1d = (unsigned char *)malloc(nd * sizeof(unsigned char));
	int64_t id = 0;
	for (int n = 0; n < ndata; n++) {
	for (int itx = 0; itx < ntx; itx++) {
	for (int it = it0; it < niter; it++) {
		d1d[id++] = (unsigned char)(255 * (d[n][itx][it] - dmin) / (dmax - dmin));
	}
	}
	}
	assert(id == nd);

	// 整数データ
	int idata[7];
	idata[0] = ndata;
	idata[1] = nx;
	idata[2] = ly;
	idata[3] = 1;
	idata[4] = ntx;
	idata[5] = 1;
	idata[6] = niter - it0;

	// 実数データ
	float fdata[4];
	fdata[0] = emin;
	fdata[1] = emax;
	fdata[2] = dmin;
	fdata[3] = dmax;

	// ファイル出力
	FILE *fp;
	if ((fp = fopen(outfile, "wb")) == NULL) {
		printf("%s open error\n", outfile);
		return;
	}

	fwrite(idata, sizeof(int),            7, fp);
	fwrite(fdata, sizeof(float),          4, fp);
	fwrite(e1d,   sizeof(unsigned char), ne, fp);  // 上限4GB?
	fwrite(d1d,   sizeof(unsigned char), nd, fp);  // 上限4GB?

	fclose(fp);

	// free
	free(e1d);
	free(d1d);
}
