/*
geometry.c
誘電率、導電率分布（セル中心）
任意個数長方形
ランダムに変化: XYZ長さ, XYZ位置, 誘電率, 導電率
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>

extern double urand(int, int64_t *);
extern void getspan(int, int, int, int, int64_t *, int *, int *);

void geometry(int ndata, int nrect, int nsmot, int64_t seed,
	const int lrect[], const float eps[], const float sig[],
	int nx, int ny, int nz, int lz,
	float ****epsr, float ****sigm, float eps0, float sig0, float eps1, float sig1, int **rect)
{
	assert((ndata > 0) && (nx > 0) && (ny > 0) && (nz > 0) && (lz > 0) && (nrect >= 0) && (nsmot >= 0));
	const int nout = lrect[2];
	//printf("%d %d %d %d %d %d %d %d\n", ndata, nrect, nsmot, nx, ny, nz, lz, nout);

	// データに関するループ
	for (int idata = 0; idata < ndata; idata++) {
		// pointer
		float ***e = epsr[idata];
		float ***s = sigm[idata];

		// 全体を背景媒質で初期化、地下・地上の2層
		for (int i = 0; i < nx; i++) {
		for (int j = 0; j < ny; j++) {
		for (int k = 0; k < nz; k++) {
			e[i][j][k] = (k < lz) ? eps1 : (k == lz) ? (eps1 + eps0) / 2 : eps0;
			s[i][j][k] = (k < lz) ? sig1 : (k == lz) ? (sig1 + sig0) / 2 : sig0;
		}
		}
		}

		// 誘電率代入
		for (int irect = 0; irect < nrect; irect++) {
			const float r1 = (float)urand(3, &seed);
			const float r2 = (float)urand(3, &seed);
			const float ep = eps[0] + (eps[1] - eps[0]) * r1;
			const float si = sig[0] + (sig[1] - sig[0]) * r2;
			int i0, i1, j0, j1, k0, k1;
			getspan(nout, nx - nout, lrect[0], lrect[1], &seed, &i0, &i1);
			getspan(nout, ny - nout, lrect[0], lrect[1], &seed, &j0, &j1);
			getspan(nout, lz - nout, lrect[0], lrect[1], &seed, &k0, &k1);
			//printf("%d %d %d %d %d %d %d %d\n", idata, irect, i0, i1, j0, j1, k0, k1);
			for (int i = i0; i < i1; i++) {
			for (int j = j0; j < j1; j++) {
			for (int k = k0; k < k1; k++) {
				e[i][j][k] = ep;
				s[i][j][k] = si;
			}
			}
			}

			// debug, 直方体（図形出力用）
			rect[idata][(6 * irect) + 0] = i0;
			rect[idata][(6 * irect) + 1] = j0;
			rect[idata][(6 * irect) + 2] = k0;
			rect[idata][(6 * irect) + 3] = i1;
			rect[idata][(6 * irect) + 4] = j1;
			rect[idata][(6 * irect) + 5] = k1;
		}

		// 内部領域平滑化(7点平均, 複数回可)
		for (int m = 0; m < nsmot; m++) {
			for (int i = nout; i < nx - nout; i++) {
			for (int j = nout; j < ny - nout; j++) {
			for (int k = nout; k < lz - nout; k++) {
				e[i][j][k] = (
				e[i][j][k] + e[i - 1][j    ][k    ] + e[i + 1][j    ][k    ]
				           + e[i    ][j - 1][k    ] + e[i    ][j + 1][k    ]
				           + e[i    ][j    ][k - 1] + e[i    ][j    ][k + 1]) / 7;
				s[i][j][k] = (
				s[i][j][k] + s[i - 1][j    ][k    ] + s[i + 1][j    ][k    ]
				           + s[i    ][j - 1][k    ] + s[i    ][j + 1][k    ]
				           + s[i    ][j    ][k - 1] + s[i    ][j    ][k + 1]) / 7;
			}
			}
			}
		}
	}
}
