/*
misc.c
fdtd3d
*/

#include <stdio.h>
#include <math.h>
#include <assert.h>

extern float C;

// メッシュ
void make_mesh(int Nx, int Ny, int Nz, float Dx, float Dy, float Dz, int lz,
	float *Dt, float *Xn, float *Yn, float *Zn,
	float *dXn, float *dYn, float *dZn, float *dXc, float *dYc, float *dZc)
{
	// 格子点
	for (int i = 0; i <= Nx; i++) {
		Xn[i] = (i - Nx / 2) * Dx;
	}
	for (int j = 0; j <= Ny; j++) {
		Yn[j] = (j - Ny / 2) * Dy;
	}
	for (int k = 0; k <= Nz; k++) {
		Zn[k] = (k - lz) * Dz;
	}
/*
	for (int i = 0; i <= Nx; i++) printf("%d %f\n", i, Xn[i]);
	for (int j = 0; j <= Ny; j++) printf("%d %f\n", j, Yn[j]);
	for (int k = 0; k <= Nz; k++) printf("%d %f\n", k, Zn[k]);
*/
	// タイムステップ
	*Dt = (float)(1 / sqrt(1 / (Dx * Dx) + 1 / (Dy * Dy) + 1 / (Dz * Dz)) / C);
	//printf("%f %f %f %e\n", Dx, Dy, Dz, *Dt);

	// C * Dt
	const float cdt = C * (*Dt);

	// 格子点に関する因子
	for (int i = 1; i < Nx; i++) {
		dXn[i] = cdt / ((Xn[i + 1] - Xn[i - 1]) / 2);
	}
	for (int j = 1; j < Ny; j++) {
		dYn[j] = cdt / ((Yn[j + 1] - Yn[j - 1]) / 2);
	}
	for (int k = 1; k < Nz; k++) {
		dZn[k] = cdt / ((Zn[k + 1] - Zn[k - 1]) / 2);
	}
	dXn[0]  = cdt / (Xn[1]  - Xn[0]);
	dYn[0]  = cdt / (Yn[1]  - Yn[0]);
	dZn[0]  = cdt / (Zn[1]  - Zn[0]);
	dXn[Nx] = cdt / (Xn[Nx] - Xn[Nx - 1]);
	dYn[Ny] = cdt / (Yn[Ny] - Yn[Ny - 1]);
	dZn[Nz] = cdt / (Zn[Nz] - Zn[Nz - 1]);

	// セル中心に関する因子
	for (int i = 0; i < Nx; i++) {
		dXc[i] = cdt / (Xn[i + 1] - Xn[i]);
	}
	for (int j = 0; j < Ny; j++) {
		dYc[j] = cdt / (Yn[j + 1] - Yn[j]);
	}
	for (int k = 0; k < Nz; k++) {
		dZc[k] = cdt / (Zn[k + 1] - Zn[k]);
	}
/*
	for (int i = 0; i <= Nx; i++) printf("%d %.5f\n", i, dXn[i]);
	for (int j = 0; j <= Ny; j++) printf("%d %.5f\n", j, dYn[j]);
	for (int k = 0; k <= Nz; k++) printf("%d %.5f\n", k, dZn[k]);
	for (int i = 0; i <  Nx; i++) printf("%d %.5f\n", i, dXc[i]);
	for (int j = 0; j <  Ny; j++) printf("%d %.5f\n", j, dYc[j]);
	for (int k = 0; k <  Nz; k++) printf("%d %.5f\n", k, dZc[k]);
*/
}


// 送受信点座標
void txrx(int Nx, int Ny, int Nz, int Lz, int Lzant,
	int NTx, int iTx0, int DTx, int **Tx,
	int NRx, int LTxRx, int DyRx, int ***Rx)
{
	// 給電点
	for (int itx = 0; itx < NTx; itx++) {
		Tx[itx][0] = iTx0 + (itx * DTx);
		Tx[itx][1] = Ny / 2;
		Tx[itx][2] = Lz + Lzant;
		//printf("%d %d %d %d\n", itx, Tx[itx][0], Tx[itx][1], Tx[itx][2]);
		assert((Tx[itx][0] > 0) && (Tx[itx][0] < Nx - 1));
		assert((Tx[itx][1] > 0) && (Tx[itx][1] < Ny - 0));
		assert((Tx[itx][2] > 0) && (Tx[itx][2] < Nz - 0));
	}

	// 観測点
	for (int itx = 0; itx < NTx; itx++) {
		for (int irx = 0; irx < NRx; irx++) {
			Rx[itx][irx][0] = Tx[itx][0] + LTxRx;
			Rx[itx][irx][1] = Tx[itx][1] + (-(NRx - 1) / 2 + irx) * DyRx;
			Rx[itx][irx][2] = Tx[itx][2];
			//printf("%d %d %d %d %d\n", itx, irx, Rx[itx][irx][0], Rx[itx][irx][1], Rx[itx][irx][2]);
			assert((Rx[itx][irx][0] > 0) && (Rx[itx][irx][0] < Nx - 1));
			assert((Rx[itx][irx][1] > 0) && (Rx[itx][irx][1] < Ny - 0));
			assert((Rx[itx][irx][2] > 0) && (Rx[itx][irx][2] < Nz - 0));
		}
	}
}


// 外部または境界のYee格子電界点から一番近い内部の2個のセル番号を求める
// (0, 1, 2)は(X, Y, Z)を巡回させたもの
void yee2cell(int n0, int n1, int n2, int i0, int i1, int i2, int *c0a, int *c0b, int *c1a, int *c1b, int *c2a, int *c2b)
{
	// 0:同じ方向
	*c0a = *c0b = (i0 < 0) ? 0 : (i0 >= n0) ? (n0 - 1) : i0;

	// 1/2:異なる方向
	*c1a = *c1b = i1;
	*c2a = *c2b = i2;
	// 4頂点外
	if      ((i1 <=  0) & (i2 <=  0)) {
		*c1a = *c1b = 0;
		*c2a = *c2b = 0;
	}
	else if ((i1 <=  0) & (i2 >= n2)) {
		*c1a = *c1b = 0;
		*c2a = *c2b = n2 - 1;
	}
	else if ((i1 >= n1) & (i2 <=  0)) {
		*c1a = *c1b = n1 - 1;
		*c2a = *c2b = 0;
	}
	else if ((i1 >= n1) & (i2 >= n2)) {
		*c1a = *c1b = n1 - 1;
		*c2a = *c2b = n2 - 1;
	}
	// 4辺外
	else if (i1 <=  0) {
		*c1a = *c1b = 0;
		*c2a = i2 - 1;
		*c2b = i2;
	}
	else if (i1 >= n1) {
		*c1a = *c1b = n1 - 1;
		*c2a = i2 - 1;
		*c2b = i2;
	}
	else if (i2 <=  0) {
		*c1a = i1 - 1;
		*c1b = i1;
		*c2a = *c2b = 0;
	}
	else if (i2 >= n2) {
		*c1a = i1 - 1;
		*c1b = i1;
		*c2a = *c2b = n2 - 1;
	}
	assert((*c0a >= 0) && (*c0a < n0));
	assert((*c0b >= 0) && (*c0b < n0));
	assert((*c1a >= 0) && (*c1a < n1));
	assert((*c1b >= 0) && (*c1b < n1));
	assert((*c2a >= 0) && (*c2a < n2));
	assert((*c2b >= 0) && (*c2b < n2));
	//assert((*j1 >= 0) && (*j1 < Ny) && (*j2 >= 0) && (*j2 < Ny));
	//assert((*k1 >= 0) && (*k1 < Nz) && (*k2 >= 0) && (*k2 < Nz));
}


// メモリーサイズ[MB]
int memory_size(int nthread, int Ndata, int Nx, int Ny, int Nz, int NTx, int NRx, int Niter, int Labc, int fig2)
{
	size_t mem = 0;
	const size_t n3d = Nx * Ny * Nz;
	const size_t n3dabc = (Nx + 2 * Labc) * (Ny + 2 * Labc) * (Nz + 2 * Labc);
	const size_t n3r = NTx * NRx * Niter;

	mem += Ndata * n3d * 2          * sizeof(float);  // Epsr, Sigm
	mem += Ndata * n3r              * sizeof(float);  // Vr
	mem += nthread * n3dabc * 2 * 3 * sizeof(float);  // PEx, PEy, PEz
	mem += nthread * n3d * 6        * sizeof(float);  // C1Ex, C2Ex, C1Ey, C2Ey, C1Ez, C2Ez
	mem += nthread * n3dabc * 6     * sizeof(float);  // Ex, Ey, Ez, Hx, Hy, Hz
	mem += nthread * n3dabc * 12    * sizeof(float);  // ABC
	mem += (fig2 > 0) * n3d * 6     * sizeof(float);  // Ex_r, Ex_i, Ey_r, Ey_i, Ez_r, Ez_i
	mem += (fig2 > 0) * NTx * n3d   * sizeof(float);  // Ea

	return (int)(mem * 1e-6) + 1;
}
