/*
mur.c
Mur-ABC
fdtd3d
*/

#include <math.h>

extern float C;

// Ex係数
void setupMurEx(int Nx, int Ny, int Nz, float *Yn, float *Zn, float Dt,
	float ****PEx, float ****Exmury, float ****Exmurz)
{
	// Y境界
	for (int j = 0; j < Ny + 1; j += Ny) {
		const float d = (j == 0) ? Yn[1] - Yn[0] : Yn[Ny] - Yn[Ny - 1];
		for (int i = 0; i < Nx + 0; i++) {
		for (int k = 0; k < Nz + 1; k++) {
			const float v = C / (float)sqrt(PEx[i][j][k][0]);
			Exmury[i][j][k][1] = (v * Dt - d) / (v * Dt + d);
		}
		}
	}

	// Z境界
	for (int k = 0; k < Nz + 1; k += Nz) {
		const float d = (k == 0) ? Zn[1] - Zn[0] : Zn[Nz] - Zn[Nz - 1];
		for (int i = 0; i < Nx + 0; i++) {
		for (int j = 0; j < Ny + 1; j++) {
			const float v = C / (float)sqrt(PEx[i][j][k][0]);
			Exmurz[i][j][k][1] = (v * Dt - d) / (v * Dt + d);
		}
		}
	}
}


// Ey係数
void setupMurEy(int Nx, int Ny, int Nz, float *Zn, float *Xn, float Dt,
	float ****PEy, float ****Eymurz, float ****Eymurx)
{
	// Z境界
	for (int k = 0; k < Nz + 1; k += Nz) {
		const float d = (k == 0) ? Zn[1] - Zn[0] : Zn[Nz] - Zn[Nz - 1];
		for (int j = 0; j < Ny + 0; j++) {
		for (int i = 0; i < Nx + 1; i++) {
			const float v = C / (float)sqrt(PEy[i][j][k][0]);
			Eymurz[i][j][k][1] = (v * Dt - d) / (v * Dt + d);
		}
		}
	}

	// X境界
	for (int i = 0; i < Nx + 1; i += Nx) {
		const float d = (i == 0) ? Xn[1] - Xn[0] : Xn[Nx] - Xn[Nx - 1];
		for (int j = 0; j < Ny + 0; j++) {
		for (int k = 0; k < Nz + 1; k++) {
			const float v = C / (float)sqrt(PEy[i][j][k][0]);
			Eymurx[i][j][k][1] = (v * Dt - d) / (v * Dt + d);
		}
		}
	}
}


// Ez係数
void setupMurEz(int Nx, int Ny, int Nz, float *Xn, float *Yn, float Dt,
	float ****PEz, float ****Ezmurx, float ****Ezmury)
{
	// X境界
	for (int i = 0; i < Nx + 1; i += Nx) {
		const float d = (i == 0) ? Xn[1] - Xn[0] : Xn[Nx] - Xn[Nx - 1];
		for (int k = 0; k < Nz + 0; k++) {
		for (int j = 0; j < Ny + 1; j++) {
			const float v = C / (float)sqrt(PEz[i][j][k][0]);
			Ezmurx[i][j][k][1] = (v * Dt - d) / (v * Dt + d);
		}
		}
	}

	// Y境界
	for (int j = 0; j < Ny + 1; j += Ny) {
		const float d = (j == 0) ? Yn[1] - Yn[0] : Yn[Ny] - Yn[Ny - 1];
		for (int k = 0; k < Nz + 0; k++) {
		for (int i = 0; i < Nx + 1; i++) {
			const float v = C / (float)sqrt(PEz[i][j][k][0]);
			Ezmury[i][j][k][1] = (v * Dt - d) / (v * Dt + d);
		}
		}
	}
}


// Ex更新
void murEx(int Nx, int Ny, int Nz, float ***Ex, float ****Exmury, float ****Exmurz)
{
	// Y境界
	for (int i = 0; i < Nx + 0; i++) {
	for (int k = 0; k < Nz + 1; k++) {
		Ex[i][ 0][k] = Exmury[i][ 0][k][0] + Exmury[i][ 0][k][1] * (Ex[i][     1][k] - Ex[i][ 0][k]);
		Ex[i][Ny][k] = Exmury[i][Ny][k][0] + Exmury[i][Ny][k][1] * (Ex[i][Ny - 1][k] - Ex[i][Ny][k]);
		Exmury[i][ 0][k][0] = Ex[i][     1][k];
		Exmury[i][Ny][k][0] = Ex[i][Ny - 1][k];
	}
	}

	// Z境界
	for (int i = 0; i < Nx + 0; i++) {
	for (int j = 0; j < Ny + 1; j++) {
		Ex[i][j][ 0] = Exmurz[i][j][ 0][0] + Exmurz[i][j][ 0][1] * (Ex[i][j][     1] - Ex[i][j][ 0]);
		Ex[i][j][Nz] = Exmurz[i][j][Nz][0] + Exmurz[i][j][Nz][1] * (Ex[i][j][Nz - 1] - Ex[i][j][Nz]);
		Exmurz[i][j][ 0][0] = Ex[i][j][     1];
		Exmurz[i][j][Nz][0] = Ex[i][j][Nz - 1];
	}
	}
}


// Ey更新
void murEy(int Nx, int Ny, int Nz, float ***Ey, float ****Eymurz, float ****Eymurx)
{
	// Z境界
	for (int j = 0; j < Ny + 0; j++) {
	for (int i = 0; i < Nx + 1; i++) {
		Ey[i][j][ 0] = Eymurz[i][j][ 0][0] + Eymurz[i][j][ 0][1] * (Ey[i][j][     1] - Ey[i][j][ 0]);
		Ey[i][j][Nz] = Eymurz[i][j][Nz][0] + Eymurz[i][j][Nz][1] * (Ey[i][j][Nz - 1] - Ey[i][j][Nz]);
		Eymurz[i][j][ 0][0] = Ey[i][j][     1];
		Eymurz[i][j][Nz][0] = Ey[i][j][Nz - 1];
	}
	}

	// X境界
	for (int j = 0; j < Ny + 0; j++) {
	for (int k = 0; k < Nz + 1; k++) {
		Ey[ 0][j][k] = Eymurx[ 0][j][k][0] + Eymurx[ 0][j][k][1] * (Ey[     1][j][k] - Ey[ 0][j][k]);
		Ey[Nx][j][k] = Eymurx[Nx][j][k][0] + Eymurx[Nx][j][k][1] * (Ey[Nx - 1][j][k] - Ey[Nx][j][k]);
		Eymurx[ 0][j][k][0] = Ey[     1][j][k];
		Eymurx[Nx][j][k][0] = Ey[Nx - 1][j][k];
	}
	}
}


// Ez更新
void murEz(int Nx, int Ny, int Nz, float ***Ez, float ****Ezmurx, float ****Ezmury)
{
	// X境界
	for (int k = 0; k < Nz + 0; k++) {
	for (int j = 0; j < Ny + 1; j++) {
		Ez[ 0][j][k] = Ezmurx[ 0][j][k][0] + Ezmurx[ 0][j][k][1] * (Ez[     1][j][k] - Ez[ 0][j][k]);
		Ez[Nx][j][k] = Ezmurx[Nx][j][k][0] + Ezmurx[Nx][j][k][1] * (Ez[Nx - 1][j][k] - Ez[Nx][j][k]);
		Ezmurx[ 0][j][k][0] = Ez[     1][j][k];
		Ezmurx[Nx][j][k][0] = Ez[Nx - 1][j][k];
	}
	}

	// Y境界
	for (int k = 0; k < Nz + 0; k++) {
	for (int i = 0; i < Nx + 1; i++) {
		Ez[i][ 0][k] = Ezmury[i][ 0][k][0] + Ezmury[i][ 0][k][1] * (Ez[i][     1][k] - Ez[i][ 0][k]);
		Ez[i][Ny][k] = Ezmury[i][Ny][k][0] + Ezmury[i][Ny][k][1] * (Ez[i][Ny - 1][k] - Ez[i][Ny][k]);
		Ezmury[i][ 0][k][0] = Ez[i][     1][k];
		Ezmury[i][Ny][k][0] = Ez[i][Ny - 1][k];
	}
	}
}
