/*
plot2d.c
fdtd3d
2D図形出力, debug用
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "../fdtd2d/ev.h"
#include "../fdtd2d/alloc.h"

extern void getminmax4d(int, int, int, int, int, int, int, int, int, float ****, float *, float *);
extern void plot2d_contour(float, float, int, int, int, int, float **, float, float, int, int);

// 誘電率分布
void plot2d_g(int ndata, int nx, int ny, int nz, int mx, int my,
	float ****d, int icolor, int grid, int ntx, int **tx, const char fn[])
{
	char str[BUFSIZ], fmt[BUFSIZ];
	const float dm = 0.3f;
	//printf("%d %d %d %d %d %d %d\n", ndata, nx, ny, nz, mx, my, ntx);
	assert((nx > 0) && (ny > 0) && (nz > 0) && (mx > 0) && (my > 0));
	if (ndata <= 0) return;

	// XY配列
	float **f = (float **)malloc(nx * sizeof(float *));
	for (int i = 0; i < nx; i++) {
		f[i] = (float *)malloc(ny * sizeof(float));
	}

	// 最小・最大
	float dmin, dmax;
	getminmax4d(0, ndata, 0, nx, 0, ny, 0, nz, 0, d, &dmin, &dmax);
	//printf("%f %f\n", dmin, dmax);

	// 開始
	const float s = 2;
	const float w = (mx * nx) + ((mx + 1) * s);
	const float h = (my * ny) + ((my + 1) * s);
	//printf("%f %f\n", w, h);
	ev2d_init(w, h);

	// 1データ=1ページ
	for (int idata = 0; idata < ndata; idata++) {
		//printf("%d %d\n", idata, ndata);
		// 改ページ
		ev2d_newPage();

		// 等高線図, 全Z面の集合
		int k = 0;
		for (int px = 0; px < mx; px++) {
		for (int py = 0; py < my; py++) {
			if (k < nz) {
				//printf("%d %d %d %d\n", idata, px, py, k);
				// 原点
				const float x0 = (px * nx) + ((px + 1) * s);
				const float y0 = (py * ny) + ((py + 1) * s);

				// XYデータ
				for (int i = 0; i < nx; i++) {
				for (int j = 0; j < ny; j++) {
					f[i][j] = d[idata][i][j][k];
					//printf("%d %d %f\n", i, j, f[i][j]);
				}
				}

				// 等高線図
				//printf("%d %d %d %f %f %d %d\n", k, px, py, x0, y0, nx, ny);
				plot2d_contour(x0, y0, 0, nx, 0, ny, f, dmin, dmax, icolor, grid);

				// データ番号
				ev2d_setColor(0, 0, 0);
				sprintf(str, "%d", k + 1);
				ev2d_drawString(x0, y0 + ny - h / 60, h / 60, str);

				// 送信点
				ev2d_setColor(0, 0, 0);
				for (int n = 0; n < ntx; n++) {
					if (k == tx[n][2]) {
						ev2d_fillRectangle(x0 + tx[n][0] - dm, y0 + tx[n][1] - dm,
						                   x0 + tx[n][0] + dm, y0 + tx[n][1] + dm);
					}
				}

				// 図++
				k++;
			}
		}
		}

		// 最小・最大
		strcpy(fmt, "data=%d/%d Nx=%d Ny=%d Nz=%d Tx=%d min=%.4f max=%.4f");
		sprintf(str, fmt, idata + 1, ndata, nx, ny, nz, ntx, dmin, dmax);
		ev2d_setColor(0, 0, 0);
		ev2d_drawString(s, s, h / 30, str);

	}

	free(f);

	// 出力
	ev2d_file(1, fn);
	ev2d_output();
}


// 電界分布
void plot2d_e(int mx, int my, int ntx, int nstep, int nx, int ny, int nz, int lz,
	float ****d, int db, int grid, int nrect, int **rect, const char fn[])
{
	char str[BUFSIZ], fmt[BUFSIZ];
	const float eps = 1e-12f;
	assert((ntx > 0) && (nstep > 0) && (mx > 0) && (my > 0) && (nx > 0) && (ny > 0) && (nz > 0) && (lz > 0));

	// 最小・最大(ZXYに注意)
	float dmin, dmax;
	getminmax4d(0, ntx, 0, ny, 0, nx, 0, nz, db, d, &dmin, &dmax);
	//printf("%f %f\n", dmin, dmax);

	// 開始
	const float s = 2;
	const float w = (mx * nx) + ((mx + 1) * s);
	const float h = (my * nz) + ((my + 1) * s);
	ev2d_init(w, h);

	// 1送信点=1ページ
	for (int itx = nstep - 1; itx < ntx; itx += nstep) {
		// 改ページ
		ev2d_newPage();

		// 等高線図, 全Y面の集合
		int j = 0;
		for (int px = 0; px < mx; px++) {
		for (int py = 0; py < my; py++) {
			if (j < ny) {
				// 原点
				const float x0 = (px * nx) + ((px + 1) * s);
				const float y0 = (py * nz) + ((py + 1) * s);
				//printf("%d %d %d %d %f %f\n", itx, j, px, py, x0, y0);

				// XY配列
				float **f = d[itx][j];

				// dB
				if (db) {
					for (int i = 0; i < nx; i++) {
					for (int k = 0; k < nz; k++) {
						f[i][k] = 20 * (float)log10(fmax(f[i][k], eps));
					}
					}
				}

				// 等高線図(カラー)
				plot2d_contour(x0, y0, 0, nx, 0, nz, f, dmin, dmax, 1, grid);

				// 地表面
				ev2d_setColor(0, 0, 0);
				ev2d_drawLine(x0, y0 + lz, x0 + nx, y0 + lz);

				// 直方体
				ev2d_setColor(0, 0, 0);
				const int idata = 0;  // 図形出力はデータ0のみ
				for (int irect = 0; irect < nrect; irect++) {
					int *ptr = &rect[idata][6 * irect];
					if ((j >= ptr[1]) && (j < ptr[4])) {
						ev2d_drawRectangle(x0 + ptr[0], y0 + ptr[2], x0 + ptr[3], y0 + ptr[5]);
					}
				}

				// データ番号
				ev2d_setColor(0, 0, 0);
				sprintf(str, "%d", j + 1);
				ev2d_drawString(x0, y0 + nz - h / 60, h / 60, str);

				// 図++
				j++;
			}
		}
		}

		// 最小・最大
		strcpy(fmt, db ? "Tx=%d/%d Nx=%d Ny=%d Nz=%d min=%.3fdB max=%.3fdB"
		               : "Tx=%d/%d Nx=%d Ny=%d Nz=%d min=%.6f max=%.6f");
		sprintf(str, fmt, itx + 1, ntx, nx, ny, nz, dmin, dmax);
		ev2d_setColor(0, 0, 0);
		ev2d_drawString(s, s, h / 40, str);
	}

	// 出力
	ev2d_file(1, fn);
	ev2d_output();
}


// レーダー画像, 線形
void plot2d_r(int ndata, int nstep, int ntx, int nrx, int niter,
	float ****r, int it0, int icolor, const char fn[])
{
	char str[BUFSIZ], fmt[BUFSIZ];
	//printf("%d %d %d %d %d %d %d\n", ndata, nstep, ntx, nrx, niter, it0, icolor);
	assert((ndata > 0) && (nstep > 0) && (ntx > 0) && (nrx > 0) && (niter > 0) && (it0 > 0));

	// 最小・最大
	float dmin, dmax;
	getminmax4d(0, ndata, 0, ntx, 0, nrx, it0, niter, 0, r, &dmin, &dmax);
	//printf("%f %f\n", dmin, dmax);

	// 最小値=-最大値, 青赤2色に適している
	if (dmin + dmax > 0) {
		dmin = -dmax;
	}
	else {
		dmax = -dmin;
	}

	// 単位図の大きさ
	const int nx = ntx;
	const int ny = niter - it0;

	// 2D配列
	float **d;
	alloc2d(float, d, nx, ny)

	// 開始
	const float s = 2;
	const float w = (nrx * nx) + ((nrx + 1) * s);
	const float h = ny + (2 * s);
	ev2d_init(w, h);

	// 1データ1ページ
	for (int idata = nstep - 1; idata < ndata; idata += nstep) {
		// 改ページ
		ev2d_newPage();

		// 図数=受信点数
		for (int irx = 0; irx < nrx; irx++) {
			// 関数値, 上下反転
			for (int i = 0; i < nx; i++) {
			for (int j = 0; j < ny; j++) {
				d[i][j] = r[idata][i][irx][niter - 1 - j];
			}
			}

			// 原点
			const float x0 = (irx * nx) + ((irx + 1) * s);
			const float y0 = s;
			//printf("%d %d %d %d %f %f\n", nx, ny, idata, irx, x0, y0);

			// 等高線図
			plot2d_contour(x0, y0, 0, nx, 0, ny, d, dmin, dmax, icolor, 0);

			// 領域
			ev2d_setColor(0, 0, 0);
			ev2d_drawRectangle(x0, y0, x0 + nx, y0 + ny);
		}

		// 最小・最大
		strcpy(fmt, "data=%d/%d Tx=%d Rx=%d time=%d-%d min=%.6f max=%.6f");
		sprintf(str, fmt, idata + 1, ndata, ntx, nrx, it0, niter, dmin, dmax);
		ev2d_drawString(s, 2 * s, h / 40, str);
	}

	// 出力
	ev2d_file(1, fn);
	ev2d_output();
}
