/*
pml.c
PML-ABC
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

extern void pmlfactor(int, float *);

extern float C;

// PML係数
void setupPML(int Nx, int Ny, int Nz, int Labc, float Dt, float *Xn, float *Yn, float *Zn,
	float *fexy, float *fexz, float *feyz, float *feyx, float *fezx, float *fezy,
	float *fhxy, float *fhxz, float *fhyz, float *fhyx, float *fhzx, float *fhzy)
{
	// PML因子
	float *f = (float *)malloc((2 * Labc + 1) * sizeof(float));
	pmlfactor(Labc, f);

	const float cdt = C * Dt;
	float fc = 0;

	// Ex
	for (int j = -Labc; j < Ny + Labc + 1; j++) {
		if      (j <      1) fc = f[- 2 * (j     )    ] / (Yn[1 ] - Yn[0     ]);
		else if (j > Ny - 1) fc = f[+ 2 * (j - Ny)    ] / (Yn[Ny] - Yn[Ny - 1]);
		else                 fc = 0;
		fexy[j + Labc] = 1 / (1 + cdt * fc);
	}
	for (int k = -Labc; k < Nz + Labc + 1; k++) {
		if      (k <      1) fc = f[- 2 * (k     )    ] / (Zn[1 ] - Zn[0     ]);
		else if (k > Nz - 1) fc = f[+ 2 * (k - Nz)    ] / (Zn[Nz] - Zn[Nz - 1]);
		else                 fc = 0;
		fexz[k + Labc] = 1 / (1 + cdt * fc);
	}

	// Ey
	for (int k = -Labc; k < Nz + Labc + 1; k++) {
		if      (k <      1) fc = f[- 2 * (k     )    ] / (Zn[1 ] - Zn[0     ]);
		else if (k > Nz - 1) fc = f[+ 2 * (k - Nz)    ] / (Zn[Nz] - Zn[Nz - 1]);
		else                 fc = 0;
		feyz[k + Labc] = 1 / (1 + cdt * fc);
	}
	for (int i = -Labc; i < Nx + Labc + 1; i++) {
		if      (i <      1) fc = f[- 2 * (i     )    ] / (Xn[1 ] - Xn[0     ]);
		else if (i > Nx - 1) fc = f[+ 2 * (i - Nx)    ] / (Xn[Nx] - Xn[Nx - 1]);
		else                 fc = 0;
		feyx[i + Labc] = 1 / (1 + cdt * fc);
	}

	// Ez
	for (int i = -Labc; i < Nx + Labc + 1; i++) {
		if      (i <      1) fc = f[- 2 * (i     )    ] / (Xn[1 ] - Xn[0     ]);
		else if (i > Nx - 1) fc = f[+ 2 * (i - Nx)    ] / (Xn[Nx] - Xn[Nx - 1]);
		else                 fc = 0;
		fezx[i + Labc] = 1 / (1 + cdt * fc);
	}
	for (int j = -Labc; j < Ny + Labc + 1; j++) {
		if      (j <      1) fc = f[- 2 * (j     )    ] / (Yn[1 ] - Yn[0     ]);
		else if (j > Ny - 1) fc = f[+ 2 * (j - Ny)    ] / (Yn[Ny] - Yn[Ny - 1]);
		else                 fc = 0;
		fezy[j + Labc] = 1 / (1 + cdt * fc);
	}

	// Hx
	for (int j = -Labc; j < Ny + Labc + 0; j++) {
		if      (j <      0) fc = f[- 2 * (j     ) - 1] / (Yn[1 ] - Yn[0     ]);
		else if (j > Ny - 1) fc = f[+ 2 * (j - Ny) + 1] / (Yn[Ny] - Yn[Ny - 1]);
		else                 fc = 0;
		fhxy[j + Labc] = 1 / (1 + cdt * fc);
	}
	for (int k = -Labc; k < Nz + Labc + 0; k++) {
		if      (k <      0) fc = f[- 2 * (k     ) - 1] / (Zn[1 ] - Zn[0     ]);
		else if (k > Nz - 1) fc = f[+ 2 * (k - Nz) + 1] / (Zn[Nz] - Zn[Nz - 1]);
		else                 fc = 0;
		fhxz[k + Labc] = 1 / (1 + cdt * fc);
	}

	// Hy
	for (int k = -Labc; k < Nz + Labc + 0; k++) {
		if      (k <      0) fc = f[- 2 * (k     ) - 1] / (Zn[1 ] - Zn[0     ]);
		else if (k > Nz - 1) fc = f[+ 2 * (k - Nz) + 1] / (Zn[Nz] - Zn[Nz - 1]);
		else                 fc = 0;
		fhyz[k + Labc] = 1 / (1 + cdt * fc);
	}
	for (int i = -Labc; i < Nx + Labc + 0; i++) {
		if      (i <      0) fc = f[- 2 * (i     ) - 1] / (Xn[1 ] - Xn[0     ]);
		else if (i > Nx - 1) fc = f[+ 2 * (i - Nx) + 1] / (Xn[Nx] - Xn[Nx - 1]);
		else                 fc = 0;
		fhyx[i + Labc] = 1 / (1 + cdt * fc);
	}

	// Hz
	for (int i = -Labc; i < Nx + Labc + 0; i++) {
		if      (i <      0) fc = f[- 2 * (i     ) - 1] / (Xn[1 ] - Xn[0     ]);
		else if (i > Nx - 1) fc = f[+ 2 * (i - Nx) + 1] / (Xn[Nx] - Xn[Nx - 1]);
		else                 fc = 0;
		fhzx[i + Labc] = 1 / (1 + cdt * fc);
	}
	for (int j = -Labc; j < Ny + Labc + 0; j++) {
		if      (j <      0) fc = f[- 2 * (j     ) - 1] / (Yn[1 ] - Yn[0     ]);
		else if (j > Ny - 1) fc = f[+ 2 * (j - Ny) + 1] / (Yn[Ny] - Yn[Ny - 1]);
		else                 fc = 0;
		fhzy[j + Labc] = 1 / (1 + cdt * fc);
	}

	// free
	free(f);
}


// Ex
void pmlEx(int Labc, int Nx, int Ny, int Nz,
	float *dYn, float *dZn, float *fexy, float *fexz,
	float ***Ex, float ***Hy, float ***Hz, float ***Exy, float ***Exz, float ****PEx)
{
	for (int i = -Labc + 0; i < Nx + Labc + 0; i++) {
	for (int j = -Labc + 1; j < Ny + Labc + 0; j++) {
	for (int k = -Labc + 1; k < Nz + Labc + 0; k++) {
		if ((i < 0) || (Nx - 1 < i) ||
		    (j < 0) || (Ny - 0 < j) ||
		    (k < 0) || (Nz - 0 < k)) {
			const int i_ = i + Labc;
			const int j_ = j + Labc;
			const int k_ = k + Labc;
			const float repsr = 1 / PEx[i_][j_][k_][0];  // 誘電率の逆数

			const float dy = dYn[(j < 0) ? 0 : (j > Ny - 0) ? Ny - 0 : j] * repsr;
			const float dhz = Hz[i_    ][j_    ][k_    ]
			                - Hz[i_    ][j_ - 1][k_    ];
			Exy[i_][j_][k_] = (
			Exy[i_][j_][k_] + dy * dhz) * fexy[j_];

			const float dz = dZn[(k < 0) ? 0 : (k > Nz - 0) ? Nz - 0 : k] * repsr;
			const float dhy = Hy[i_    ][j_    ][k_    ]
			                - Hy[i_    ][j_    ][k_ - 1];
			Exz[i_][j_][k_] = (
			Exz[i_][j_][k_] - dz * dhy) * fexz[k_];

			Ex[i_][j_][k_] = Exy[i_][j_][k_] + Exz[i_][j_][k_];
		}
	}
	}
	}
}


// Ey
void pmlEy(int Labc, int Nx, int Ny, int Nz,
	float *dZn, float *dXn, float *feyz, float *feyx,
	float ***Ey, float ***Hz, float ***Hx, float ***Eyz, float ***Eyx, float ****PEy)
{
	for (int i = -Labc + 1; i < Nx + Labc + 0; i++) {
	for (int j = -Labc + 0; j < Ny + Labc + 0; j++) {
	for (int k = -Labc + 1; k < Nz + Labc + 0; k++) {
		if ((i < 0) || (Nx - 0 < i) ||
		    (j < 0) || (Ny - 1 < j) ||
		    (k < 0) || (Nz - 0 < k)) {
			const int i_ = i + Labc;
			const int j_ = j + Labc;
			const int k_ = k + Labc;
			const float repsr = 1 / PEy[i_][j_][k_][0];  // 誘電率の逆数

			const float dz = dZn[(k < 0) ? 0 : (k > Nz - 0) ? Nz - 0 : k] * repsr;
			const float dhx = Hx[i_    ][j_    ][k_    ]
			                - Hx[i_    ][j_    ][k_ - 1];
			Eyz[i_][j_][k_] = (
			Eyz[i_][j_][k_] + dz * dhx) * feyz[k_];

			const float dx = dXn[(i < 0) ? 0 : (i > Nx - 0) ? Nx - 0 : i] * repsr;
			const float dhz = Hz[i_    ][j_    ][k_    ]
			                - Hz[i_ - 1][j_    ][k_    ];
			Eyx[i_][j_][k_] = (
			Eyx[i_][j_][k_] - dx * dhz) * feyx[i_];

			Ey[i_][j_][k_] = Eyz[i_][j_][k_] + Eyx[i_][j_][k_];
		}
	}
	}
	}
}


// Ez
void pmlEz(int Labc, int Nx, int Ny, int Nz,
	float *dXn, float *dYn, float *fezx, float *fezy,
	float ***Ez, float ***Hx, float ***Hy, float ***Ezx, float ***Ezy, float ****PEz)
{
	for (int i = -Labc + 1; i < Nx + Labc + 0; i++) {
	for (int j = -Labc + 1; j < Ny + Labc + 0; j++) {
	for (int k = -Labc + 0; k < Nz + Labc + 0; k++) {
		if ((i < 0) || (Nx - 0 < i) ||
		    (j < 0) || (Ny - 0 < j) ||
		    (k < 0) || (Nz - 1 < k)) {
			const int i_ = i + Labc;
			const int j_ = j + Labc;
			const int k_ = k + Labc;
			const float repsr = 1 / PEz[i_][j_][k_][0];  // 誘電率の逆数

			const float dx = dXn[(i < 0) ? 0 : (i > Nx - 0) ? Nx - 0 : i] * repsr;
			const float dhy = Hy[i_    ][j_    ][k_    ]
			                - Hy[i_ - 1][j_    ][k_    ];
			Ezx[i_][j_][k_] = (
			Ezx[i_][j_][k_] + dx * dhy) * fezx[i_];

			const float dy = dYn[(j < 0) ? 0 : (j > Ny - 0) ? Ny - 0 : j] * repsr;
			const float dhx = Hx[i_    ][j_    ][k_    ]
			                - Hx[i_    ][j_ - 1][k_    ];
			Ezy[i_][j_][k_] = (
			Ezy[i_][j_][k_] - dy * dhx) * fezy[j_];

			Ez[i_][j_][k_] = Ezx[i_][j_][k_] + Ezy[i_][j_][k_];
		}
	}
	}
	}
}


// Hx
void pmlHx(int Labc, int Nx, int Ny, int Nz,
	float *dYc, float *dZc, float *fhxy, float *fhxz,
	float ***Hx, float ***Ey, float ***Ez, float ***Hxy, float ***Hxz)
{
	const float ramur = 1;  // 磁性体なし

	for (int i = -Labc + 1; i < Nx + Labc + 0; i++) {
	for (int j = -Labc + 0; j < Ny + Labc + 0; j++) {
	for (int k = -Labc + 0; k < Nz + Labc + 0; k++) {
		if ((i < 0) || (Nx - 0 < i) ||
		    (j < 0) || (Ny - 1 < j) ||
		    (k < 0) || (Nz - 1 < k)) {
			const int i_ = i + Labc;
			const int j_ = j + Labc;
			const int k_ = k + Labc;

			const float dy = dYc[(j < 0) ? 0 : (j > Ny - 1) ? Ny - 1 : j] * ramur;
			const float dez = Ez[i_    ][j_ + 1][k_    ]
			                - Ez[i_    ][j_    ][k_    ];
			Hxy[i_][j_][k_] = (
			Hxy[i_][j_][k_] - (dy * dez)) * fhxy[j_];

			const float dz = dZc[(k < 0) ? 0 : (k > Nz - 1) ? Nz - 1 : k] * ramur;
			const float dey = Ey[i_    ][j_    ][k_ + 1]
			                - Ey[i_    ][j_    ][k_    ];
			Hxz[i_][j_][k_] = (
			Hxz[i_][j_][k_] + (dz * dey)) * fhxz[k_];

			Hx[i_][j_][k_] = Hxy[i_][j_][k_] + Hxz[i_][j_][k_];
		}
	}
	}
	}
}


// Hy
void pmlHy(int Labc, int Nx, int Ny, int Nz,
	float *dZc, float *dXc, float *fhyz, float *fhyx,
	float ***Hy, float ***Ez, float ***Ex, float ***Hyz, float ***Hyx)
{
	const float ramur = 1;  // 磁性体なし

	for (int i = -Labc + 0; i < Nx + Labc + 0; i++) {
	for (int j = -Labc + 1; j < Ny + Labc + 0; j++) {
	for (int k = -Labc + 0; k < Nz + Labc + 0; k++) {
		if ((i < 0) || (Nx - 1 < i) ||
		    (j < 0) || (Ny - 0 < j) ||
		    (k < 0) || (Nz - 1 < k)) {
			const int i_ = i + Labc;
			const int j_ = j + Labc;
			const int k_ = k + Labc;

			const float dz = dZc[(k < 0) ? 0 : (k > Nz - 1) ? Nz - 1 : k] * ramur;
			const float dex = Ex[i_    ][j_    ][k_ + 1]
			                - Ex[i_    ][j_    ][k_    ];
			Hyz[i_][j_][k_] = (
			Hyz[i_][j_][k_] - (dz * dex)) * fhyz[k_];

			const float dx = dXc[(i < 0) ? 0 : (i > Nx - 1) ? Nx - 1 : i] * ramur;
			const float dez = Ez[i_ + 1][j_    ][k_    ]
			                - Ez[i_    ][j_    ][k_    ];
			Hyx[i_][j_][k_] = (
			Hyx[i_][j_][k_] + (dx * dez)) * fhyx[i_];

			Hy[i_][j_][k_] = Hyz[i_][j_][k_] + Hyx[i_][j_][k_];
		}
	}
	}
	}
}


// Hz
void pmlHz(int Labc, int Nx, int Ny, int Nz,
	float *dXc, float *dYc, float *fhzx, float *fhzy,
	float ***Hz, float ***Ex, float ***Ey, float ***Hzx, float ***Hzy)
{
	const float ramur = 1;  // 磁性体なし

	for (int i = -Labc + 0; i < Nx + Labc + 0; i++) {
	for (int j = -Labc + 0; j < Ny + Labc + 0; j++) {
	for (int k = -Labc + 1; k < Nz + Labc + 0; k++) {
		if ((i < 0) || (Nx - 1 < i) ||
		    (j < 0) || (Ny - 1 < j) ||
		    (k < 0) || (Nz - 0 < k)) {
			const int i_ = i + Labc;
			const int j_ = j + Labc;
			const int k_ = k + Labc;

			const float dx = dXc[(i < 0) ? 0 : (i > Nx - 1) ? Nx - 1 : i] * ramur;
			const float dey = Ey[i_ + 1][j_    ][k_    ]
			                - Ey[i_    ][j_    ][k_    ];
			Hzx[i_][j_][k_] = (
			Hzx[i_][j_][k_] - (dx * dey)) * fhzx[i_];

			const float dy = dYc[(j < 0) ? 0 : (j > Ny - 1) ? Ny - 1 : j] * ramur;
			const float dex = Ex[i_    ][j_ + 1][k_    ]
			                - Ex[i_    ][j_    ][k_    ];
			Hzy[i_][j_][k_] = (
			Hzy[i_][j_][k_] + (dy * dex)) * fhzy[j_];

			Hz[i_][j_][k_] = Hzx[i_][j_][k_] + Hzy[i_][j_][k_];
		}
	}
	}
	}
}
