/*
solve.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "../fdtd2d/alloc.h"

static void setupYeeEx(int, int, int, int, float ***, float ***, float ****);
static void setupYeeEy(int, int, int, int, float ***, float ***, float ****);
static void setupYeeEz(int, int, int, int, float ***, float ***, float ****);
static void setupFactor(int, int, int, int, float, float ****, float ****, float ****, float ***, float ***, float ***, float ***, float ***, float ***);
static void initfield(int, int, int, int, int, int, float ***, float ***, float ***, float ***, float ***, float ***, float **, float ****, float ****, float ****, float ****, float ****, float ****, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, float ***, int, float ***, float ***, float ***, float ***, float ***, float ***);
static void average(int, int, int, int, float ***, float ***, float ***, float ***, float ***, float ***, float []);
static void dft(int, int, int, int, float ***, float ***, float ***, float, float, float ***, float ***, float ***, float ***, float ***, float ***);

extern void dftfactor(int, float, float, int, float *, float *);
extern void updateEx(int, int, int, int, float ***, float ***, float ***, float *, float *, float ***, float ***);
extern void updateEy(int, int, int, int, float ***, float ***, float ***, float *, float *, float ***, float ***);
extern void updateEz(int, int, int, int, float ***, float ***, float ***, float *, float *, float ***, float ***);
extern void updateHx(int, int, int, int, float ***, float ***, float ***, float *, float *);
extern void updateHy(int, int, int, int, float ***, float ***, float ***, float *, float *);
extern void updateHz(int, int, int, int, float ***, float ***, float ***, float *, float *);
extern void setupMurEx(int, int, int, float *, float *, float, float ****, float ****, float ****);
extern void setupMurEy(int, int, int, float *, float *, float, float ****, float ****, float ****);
extern void setupMurEz(int, int, int, float *, float *, float, float ****, float ****, float ****);
extern void setupPML(int, int, int, int, float, float *, float *, float *, float *, float *, float *, float *, float *, float *, float *, float *, float *, float *, float *, float *);
extern void murEx(int, int, int, float ***, float ****, float ****);
extern void murEy(int, int, int, float ***, float ****, float ****);
extern void murEz(int, int, int, float ***, float ****, float ****);
extern void pmlEx(int, int, int, int, float *, float *, float *, float *, float ***, float ***, float ***, float ***, float ***, float ****);
extern void pmlEy(int, int, int, int, float *, float *, float *, float *, float ***, float ***, float ***, float ***, float ***, float ****);
extern void pmlEz(int, int, int, int, float *, float *, float *, float *, float ***, float ***, float ***, float ***, float ***, float ****);
extern void pmlHx(int, int, int, int, float *, float *, float *, float *, float ***, float ***, float ***, float ***, float ***);
extern void pmlHy(int, int, int, int, float *, float *, float *, float *, float ***, float ***, float ***, float ***, float ***);
extern void pmlHz(int, int, int, int, float *, float *, float *, float *, float ***, float ***, float ***, float ***, float ***);
extern float vfeed(float, float, float, int);
extern void yee2cell(int, int, int, int, int, int, int *, int *, int *, int *, int *, int *);

extern float C;

void solve(
	int Niter, int Nx, int Ny, int Nz, float Freq, float Dt, int Labc, int WF,
	int *Tx, int NRx, int **Rx,
	float *Xn, float *Yn, float *Zn, float *dXn, float *dYn, float *dZn, float *dXc, float *dYc, float *dZc,
	float ***epsr, float ***sigm, float ****PEx, float ****PEy, float ****PEz,
	float ***Ex, float ***Ey, float ***Ez, float ***Hx, float ***Hy, float ***Hz,
	float ***C1Ex, float ***C2Ex, float ***C1Ey, float ***C2Ey, float ***C1Ez, float ***C2Ez,
	float ****Exmury, float ****Exmurz, float ****Eymurz, float ****Eymurx, float ****Ezmurx, float ****Ezmury,
	float ***Exy, float ***Exz, float ***Eyz, float ***Eyx, float ***Ezx, float ***Ezy,
	float ***Hxy, float ***Hxz, float ***Hyz, float ***Hyx, float ***Hzx, float ***Hzy,
	float **Vr,
	int ilog, int iout, float ***Ex_r, float ***Ex_i, float ***Ey_r, float ***Ey_i, float ***Ez_r, float ***Ez_i, float ***Ea)
{
	//printf("%d %d %d %d %e %e %d\n", Niter, Nx, Ny, Nz, Freq, Dt, Labc);
	//printf("%d %d %d\n", iTx, jTx, kTx);

	// Yee格子物性値
	setupYeeEx(Nx, Ny, Nz, Labc, epsr, sigm, PEx);
	setupYeeEy(Nx, Ny, Nz, Labc, epsr, sigm, PEy);
	setupYeeEz(Nx, Ny, Nz, Labc, epsr, sigm, PEz);

	// FDTD係数
	setupFactor(Nx, Ny, Nz, Labc, Dt, PEx, PEy, PEz, C1Ex, C2Ex, C1Ey, C2Ey, C1Ez, C2Ez);

	// 電磁界初期化
	initfield(Niter, Nx, Ny, Nz, NRx, Labc,
		Ex, Ey, Ez, Hx, Hy, Hz,
		Vr,
		Exmury, Exmurz, Eymurz, Eymurx, Ezmurx, Ezmury,
		Exy, Exz, Eyz, Eyx, Ezx, Ezy,
		Hxy, Hxz, Hyz, Hyx, Hzx, Hzy,
		iout, Ex_r, Ex_i, Ey_r, Ey_i, Ez_r, Ez_i);

	// ABC係数
	float *fexy = NULL, *fexz = NULL, *feyz = NULL, *feyx = NULL, *fezx = NULL, *fezy = NULL;
	float *fhxy = NULL, *fhxz = NULL, *fhyz = NULL, *fhyx = NULL, *fhzx = NULL, *fhzy = NULL;
	if (Labc > 0) {
		alloc1d(float, fexy, Ny + 1 + 2 * Labc)
		alloc1d(float, fexz, Nz + 1 + 2 * Labc)
		alloc1d(float, feyz, Nz + 1 + 2 * Labc)
		alloc1d(float, feyx, Nx + 1 + 2 * Labc)
		alloc1d(float, fezx, Nx + 1 + 2 * Labc)
		alloc1d(float, fezy, Ny + 1 + 2 * Labc)
		alloc1d(float, fhxy, Ny + 0 + 2 * Labc)
		alloc1d(float, fhxz, Nz + 0 + 2 * Labc)
		alloc1d(float, fhyz, Nz + 0 + 2 * Labc)
		alloc1d(float, fhyx, Nx + 0 + 2 * Labc)
		alloc1d(float, fhzx, Nx + 0 + 2 * Labc)
		alloc1d(float, fhzy, Ny + 0 + 2 * Labc)
		setupPML(Nx, Ny, Nz, Labc, Dt, Xn, Yn, Zn,
			fexy, fexz, feyz, feyx, fezx, fezy,
			fhxy, fhxz, fhyz, fhyx, fhzx, fhzy);
	}
	else {
		setupMurEx(Nx, Ny, Nz, Yn, Zn, Dt, PEx, Exmury, Exmurz);
		setupMurEy(Nx, Ny, Nz, Zn, Xn, Dt, PEy, Eymurz, Eymurx);
		setupMurEz(Nx, Ny, Nz, Xn, Yn, Dt, PEz, Ezmurx, Ezmury);
	}

	// DFT係数, debug
	float *fcos = NULL, *fsin = NULL;
	if (iout) {
		alloc1d(float, fcos, Niter + 1)
		alloc1d(float, fsin, Niter + 1)
		dftfactor(Niter, Freq, Dt, WF, fcos, fsin);
	}

	// FDTD反復計算
	float t = 0;
	for (int iter = 0; iter <= Niter; iter++) {

		// E更新
		updateEx(Labc, Nx, Ny, Nz, Ex, Hy, Hz, dYn, dZn, C1Ex, C2Ex);
		updateEy(Labc, Nx, Ny, Nz, Ey, Hz, Hx, dZn, dXn, C1Ey, C2Ey);
		updateEz(Labc, Nx, Ny, Nz, Ez, Hx, Hy, dXn, dYn, C1Ez, C2Ez);

		// 給電点(X方向)
		const int itx = Tx[0];
		const int jtx = Tx[1];
		const int ktx = Tx[2];
		const float dxtx = Xn[itx + 1] - Xn[itx];
		const float v = 1000;  // 1でもよいが数値を大きくするため
		Ex[itx + Labc][jtx + Labc][ktx + Labc] = v * vfeed(t, 0, Freq, WF) / dxtx;

		// ABC
		if (Labc > 0) {
			pmlEx(Labc, Nx, Ny, Nz, dYn, dZn, fexy, fexz, Ex, Hy, Hz, Exy, Exz, PEx);
			pmlEy(Labc, Nx, Ny, Nz, dZn, dXn, feyz, feyx, Ey, Hz, Hx, Eyz, Eyx, PEy);
			pmlEz(Labc, Nx, Ny, Nz, dXn, dYn, fezx, fezy, Ez, Hx, Hy, Ezx, Ezy, PEz);
		}
		else {
			murEx(Nx, Ny, Nz, Ex, Exmury, Exmurz);
			murEy(Nx, Ny, Nz, Ey, Eymurz, Eymurx);
			murEz(Nx, Ny, Nz, Ez, Ezmurx, Ezmury);
		}

		// H更新
		updateHx(Labc, Nx, Ny, Nz, Hx, Ey, Ez, dYc, dZc);
		updateHy(Labc, Nx, Ny, Nz, Hy, Ez, Ex, dZc, dXc);
		updateHz(Labc, Nx, Ny, Nz, Hz, Ex, Ey, dXc, dYc);

		// Labc
		if (Labc > 0) {
			pmlHx(Labc, Nx, Ny, Nz, dYc, dZc, fhxy, fhxz, Hx, Ey, Ez, Hxy, Hxz);
			pmlHy(Labc, Nx, Ny, Nz, dZc, dXc, fhyz, fhyx, Hy, Ez, Ex, Hyz, Hyx);
			pmlHz(Labc, Nx, Ny, Nz, dXc, dYc, fhzx, fhzy, Hz, Ex, Ey, Hzx, Hzy);
		}

		// 受信波形（Ex成分）
		for (int rx = 0; rx < NRx; rx++) {
			const int irx = Rx[rx][0];
			const int jrx = Rx[rx][1];
			const int krx = Rx[rx][2];
			const float dxrx = Xn[irx + 1] - Xn[irx];
			Vr[rx][iter] = Ex[irx + Labc][jrx + Labc][krx + Labc] * dxrx;
			//printf("%d %d %e\n", iter, rx, Vr[rx][iter]);
		}

		// DFT, debug
		if (iout) {
			dft(Labc, Nx, Ny, Nz, Ex, Ey, Ez, fcos[iter], fsin[iter], Ex_r, Ex_i, Ey_r, Ey_i, Ez_r, Ez_i);
		}

		// 時間ステップ
		t += Dt;

		// debug, 平均電磁界出力
		if (ilog && (iter % 50 == 0)) {
			float fsum[2];
			average(Labc, Nx, Ny, Nz, Ex, Ey, Ez, Hx, Hy, Hz, fsum);
			printf("%5d %.6f %.6f\n", iter, fsum[0], fsum[1]);
			fflush(stdout);
		}
	}

	// debug, 電界分布, セル中心, データ0=プロセス0=通信不要
	if (iout) {
		for (int i = 0; i < Nx; i++) {
		for (int j = 0; j < Ny; j++) {
		for (int k = 0; k < Nz; k++) {
			// セル平均(YXZに注意)
			const float ex_r = (Ex_r[i][j][k] + Ex_r[i    ][j + 1][k    ] + Ex_r[i    ][j    ][k + 1] + Ex_r[i    ][j + 1][k + 1]) / 4;
			const float ex_i = (Ex_i[i][j][k] + Ex_i[i    ][j + 1][k    ] + Ex_i[i    ][j    ][k + 1] + Ex_i[i    ][j + 1][k + 1]) / 4;
			const float ey_r = (Ey_r[i][j][k] + Ey_r[i    ][j    ][k + 1] + Ey_r[i + 1][j    ][k    ] + Ey_r[i + 1][j    ][k + 1]) / 4;
			const float ey_i = (Ey_i[i][j][k] + Ey_i[i    ][j    ][k + 1] + Ey_i[i + 1][j    ][k    ] + Ey_i[i + 1][j    ][k + 1]) / 4;
			const float ez_r = (Ez_r[i][j][k] + Ez_r[i + 1][j    ][k    ] + Ez_r[i    ][j + 1][k    ] + Ez_r[i + 1][j + 1][k    ]) / 4;
			const float ez_i = (Ez_i[i][j][k] + Ez_i[i + 1][j    ][k    ] + Ez_i[i    ][j + 1][k    ] + Ez_i[i + 1][j + 1][k    ]) / 4;
			Ea[j][i][k] = (float)sqrt((ex_r * ex_r) + (ex_i * ex_i) + (ey_r * ey_r) + (ey_i * ey_i) + (ez_r * ez_r) + (ez_i * ez_i));
			//printf("%d %d %d %e\n", k, i, j, Ea[j][i][k]);
		}
		}
		}
	}

	// free
	if (Labc > 0) {
		free(fexy); free(fexz);
		free(feyz); free(feyx);
		free(fezx); free(fezy);
		free(fhxy); free(fhxz);
		free(fhyz); free(fhyx);
		free(fhzx); free(fhzy);
	}
	if (iout) {
		free(fcos);
		free(fsin);
	}
}


// Ex, Yee格子物性値
static void setupYeeEx(int Nx, int Ny, int Nz, int Labc, float ***epsr, float ***sigm, float ****PEx)
{
	//printf("%d %d %d %d\n", Nx, Ny, Nz, Labc);
	for (int i = -Labc; i < Nx + 0 + Labc; i++) {
	for (int j = -Labc; j < Ny + 1 + Labc; j++) {
	for (int k = -Labc; k < Nz + 1 + Labc; k++) {
		float eps, sig;
		if ((i >= 0) && (i < Nx) && (j > 0) && (j < Ny) && (k > 0) && (k < Nz)) {
			// 内部:4点の平均
			eps = (epsr[i][j - 1][k - 1] + epsr[i][j][k - 1] + epsr[i][j - 1][k] + epsr[i][j][k]) / 4;
			sig = (sigm[i][j - 1][k - 1] + sigm[i][j][k - 1] + sigm[i][j - 1][k] + sigm[i][j][k]) / 4;
		}
		else {
			// 外部:1点または2点の平均
			int i1, i2, j1, j2, k1, k2;
			yee2cell(Nx, Ny, Nz, i, j, k, &i1, &i2, &j1, &j2, &k1, &k2);
			eps = (epsr[i1][j1][k1] + epsr[i2][j2][k2]) / 2;
			sig = (sigm[i1][j1][k1] + sigm[i2][j2][k2]) / 2;
		}
		PEx[i + Labc][j + Labc][k + Labc][0] = eps;
		PEx[i + Labc][j + Labc][k + Labc][1] = sig;
		//printf("%d %d %d %e %e\n", i, j, k, PEx[i + Labc][j + Labc][k + Labc][0], PEx[i + Labc][j + Labc][k + Labc][1]);
	}
	}
	}
}


// Ey, Yee格子物性値
static void setupYeeEy(int Nx, int Ny, int Nz, int Labc, float ***epsr, float ***sigm, float ****PEy)
{
	for (int i = -Labc; i < Nx + 1 + Labc; i++) {
	for (int j = -Labc; j < Ny + 0 + Labc; j++) {
	for (int k = -Labc; k < Nz + 1 + Labc; k++) {
		float eps, sig;
		if ((j >= 0) && (j < Ny) && (k > 0) && (k < Nz) && (i > 0) && (i < Nx)) {
			// 内部:4点の平均
			eps = (epsr[i - 1][j][k - 1] + epsr[i - 1][j][k] + epsr[i][j][k - 1] + epsr[i][j][k]) / 4;
			sig = (sigm[i - 1][j][k - 1] + sigm[i - 1][j][k] + sigm[i][j][k - 1] + sigm[i][j][k]) / 4;
		}
		else {
			// 外部:1点または2点の平均
			int i1, i2, j1, j2, k1, k2;
			yee2cell(Ny, Nz, Nx, j, k, i, &j1, &j2, &k1, &k2, &i1, &i2);
			eps = (epsr[i1][j1][k1] + epsr[i2][j2][k2]) / 2;
			sig = (sigm[i1][j1][k1] + sigm[i2][j2][k2]) / 2;
		}
		PEy[i + Labc][j + Labc][k + Labc][0] = eps;
		PEy[i + Labc][j + Labc][k + Labc][1] = sig;
		//printf("%d %d %d %e %e\n", i, j, k, PEy[i + Labc][j + Labc][k + Labc][0], PEy[i + Labc][j + Labc][k + Labc][1]);
	}
	}
	}
}


// Ez, Yee格子物性値
static void setupYeeEz(int Nx, int Ny, int Nz, int Labc, float ***epsr, float ***sigm, float ****PEz)
{
	for (int i = -Labc; i < Nx + 1 + Labc; i++) {
	for (int j = -Labc; j < Ny + 1 + Labc; j++) {
	for (int k = -Labc; k < Nz + 0 + Labc; k++) {
		float eps, sig;
		if ((k >= 0) && (k < Nz) && (i > 0) && (i < Nx) && (j > 0) && (j < Ny)) {
			// 内部:4点の平均
			eps = (epsr[i - 1][j - 1][k] + epsr[i][j - 1][k] + epsr[i - 1][j][k] + epsr[i][j][k]) / 4;
			sig = (sigm[i - 1][j - 1][k] + sigm[i][j - 1][k] + sigm[i - 1][j][k] + sigm[i][j][k]) / 4;
		}
		else {
			// 外部:1点または2点の平均
			int i1, i2, j1, j2, k1, k2;
			yee2cell(Nz, Nx, Ny, k, i, j, &k1, &k2, &i1, &i2, &j1, &j2);
			eps = (epsr[i1][j1][k1] + epsr[i2][j2][k2]) / 2;
			sig = (sigm[i1][j1][k1] + sigm[i2][j2][k2]) / 2;
		}
		PEz[i + Labc][j + Labc][k + Labc][0] = eps;
		PEz[i + Labc][j + Labc][k + Labc][1] = sig;
		//printf("%d %d %d %e %e\n", i, j, k, PEz[i + Labc][j + Labc][k + Labc][0], PEz[i + Labc][j + Labc][k + Labc][1]);
	}
	}
	}
}


// FDTD係数
static void setupFactor(int Nx, int Ny, int Nz, int Labc, float Dt,
	float ****PEx, float ****PEy, float ****PEz,
	float ***C1Ex, float ***C2Ex, float ***C1Ey, float ***C2Ey, float ***C1Ez, float ***C2Ez)
{
	const float pi = (float)(4 * atan(1));
	const float eta = C * 4 * pi * 1e-7f;
	//printf("%d %d %d %d %e\n", Nx, Ny, Nz, Labc, Dt);

	// Ex
	for (int i = 0; i < Nx + 0; i++) {
	for (int j = 0; j < Ny + 1; j++) {
	for (int k = 0; k < Nz + 1; k++) {
		const float eps = PEx[i + Labc][j + Labc][k + Labc][0];
		const float sig = PEx[i + Labc][j + Labc][k + Labc][1];
		C2Ex[i][j][k] = 1 / (eps + (sig * eta * C * Dt));
		C1Ex[i][j][k] = C2Ex[i][j][k] * eps;
		//printf("%d %d %d %f %f %f %f\n", i, j, k, eps, sig, C1Ex[i][j][k], C2Ex[i][j][k]);
	}
	}
	}

	// Ey
	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 0; j++) {
	for (int k = 0; k < Nz + 1; k++) {
		const float eps = PEy[i + Labc][j + Labc][k + Labc][0];
		const float sig = PEy[i + Labc][j + Labc][k + Labc][1];
		C2Ey[i][j][k] = 1 / (eps + (sig * eta * C * Dt));
		C1Ey[i][j][k] = C2Ey[i][j][k] * eps;
	}
	}
	}

	// Ez
	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 1; j++) {
	for (int k = 0; k < Nz + 0; k++) {
		const float eps = PEz[i + Labc][j + Labc][k + Labc][0];
		const float sig = PEz[i + Labc][j + Labc][k + Labc][1];
		C2Ez[i][j][k] = 1 / (eps + (sig * eta * C * Dt));
		C1Ez[i][j][k] = C2Ez[i][j][k] * eps;
	}
	}
	}
}


// 電磁界初期化
static void initfield(int Niter, int Nx, int Ny, int Nz, int NRx, int Labc,
	float ***Ex, float ***Ey, float ***Ez, float ***Hx, float ***Hy, float ***Hz,
	float **Vr,
	float ****Exmury, float ****Exmurz, float ****Eymurz, float ****Eymurx, float ****Ezmurx, float ****Ezmury,
	float ***Exy, float ***Exz, float ***Eyz, float ***Eyx, float ***Ezx, float ***Ezy,
	float ***Hxy, float ***Hxz, float ***Hyz, float ***Hyx, float ***Hzx, float ***Hzy,
	int iout, float ***Ex_r, float ***Ex_i, float ***Ey_r, float ***Ey_i, float ***Ez_r, float ***Ez_i)
{
	assert((Niter > 0) && (Nx > 0) && (Ny > 0) && (Nz > 0) && (Labc >= 0));

	// E
	for (int i = - Labc; i < Nx + 0 + Labc; i++) {
	for (int j = - Labc; j < Ny + 1 + Labc; j++) {
		memset(Ex[i + Labc][j + Labc], 0, (Nz + 1 + 2 * Labc) * sizeof(float));
	}
	}
	for (int i = - Labc; i < Nx + 1 + Labc; i++) {
	for (int j = - Labc; j < Ny + 0 + Labc; j++) {
		memset(Ey[i + Labc][j + Labc], 0, (Nz + 1 + 2 * Labc) * sizeof(float));
	}
	}
	for (int i = - Labc; i < Nx + 1 + Labc; i++) {
	for (int j = - Labc; j < Ny + 1 + Labc; j++) {
		memset(Ez[i + Labc][j + Labc], 0, (Nz + 0 + 2 * Labc) * sizeof(float));
	}
	}
	// H
	for (int i = - Labc; i < Nx + 1 + Labc; i++) {
	for (int j = - Labc; j < Ny + 0 + Labc; j++) {
		memset(Hx[i + Labc][j + Labc], 0, (Nz + 0 + 2 * Labc) * sizeof(float));
	}
	}
	for (int i = - Labc; i < Nx + 0 + Labc; i++) {
	for (int j = - Labc; j < Ny + 1 + Labc; j++) {
		memset(Hy[i + Labc][j + Labc], 0, (Nz + 0 + 2 * Labc) * sizeof(float));
	}
	}
	for (int i = - Labc; i < Nx + 0 + Labc; i++) {
	for (int j = - Labc; j < Ny + 0 + Labc; j++) {
		memset(Hz[i + Labc][j + Labc], 0, (Nz + 1 + 2 * Labc) * sizeof(float));
	}
	}

	// 受信波形
	for (int irx = 0; irx < NRx; irx++) {
		memset(Vr[irx], 0, Niter * sizeof(float));
	}

	// ABC
	if (Labc == 0) {
		// Ex
		for (int i = 0; i < Nx + 0; i++) {
		for (int j = 0; j < Ny + 1; j++) {
		for (int k = 0; k < Nz + 1; k++) {
			Exmury[i][j][k][0] = Exmury[i][j][k][1] =
			Exmurz[i][j][k][0] = Exmurz[i][j][k][1] = 0;
		}
		}
		}
		// Ey
		for (int i = 0; i < Nx + 1; i++) {
		for (int j = 0; j < Ny + 0; j++) {
		for (int k = 0; k < Nz + 1; k++) {
			Eymurz[i][j][k][0] = Eymurz[i][j][k][1] =
			Eymurx[i][j][k][0] = Eymurx[i][j][k][1] = 0;
		}
		}
		}
		// Ez
		for (int i = 0; i < Nx + 1; i++) {
		for (int j = 0; j < Ny + 1; j++) {
		for (int k = 0; k < Nz + 0; k++) {
			Ezmurx[i][j][k][0] = Ezmurx[i][j][k][1] =
			Ezmury[i][j][k][0] = Ezmury[i][j][k][1] = 0;
		}
		}
		}
	}
	else {
		// E
		for (int i = - Labc; i < Nx + 0 + Labc; i++) {
		for (int j = - Labc; j < Ny + 1 + Labc; j++) {
			memset(Exy[i + Labc][j + Labc], 0, (Nz + 1 + 2 * Labc) * sizeof(float));
			memset(Exz[i + Labc][j + Labc], 0, (Nz + 1 + 2 * Labc) * sizeof(float));
		}
		}
		for (int i = - Labc; i < Nx + 1 + Labc; i++) {
		for (int j = - Labc; j < Ny + 0 + Labc; j++) {
			memset(Eyz[i + Labc][j + Labc], 0, (Nz + 1 + 2 * Labc) * sizeof(float));
			memset(Eyx[i + Labc][j + Labc], 0, (Nz + 1 + 2 * Labc) * sizeof(float));
		}
		}
		for (int i = - Labc; i < Nx + 1 + Labc; i++) {
		for (int j = - Labc; j < Ny + 1 + Labc; j++) {
			memset(Ezx[i + Labc][j + Labc], 0, (Nz + 0 + 2 * Labc) * sizeof(float));
			memset(Ezy[i + Labc][j + Labc], 0, (Nz + 0 + 2 * Labc) * sizeof(float));
		}
		}
		// H
		for (int i = - Labc; i < Nx + 1 + Labc; i++) {
		for (int j = - Labc; j < Ny + 0 + Labc; j++) {
			memset(Hxy[i + Labc][j + Labc], 0, (Nz + 0 + 2 * Labc) * sizeof(float));
			memset(Hxz[i + Labc][j + Labc], 0, (Nz + 0 + 2 * Labc) * sizeof(float));
		}
		}
		for (int i = - Labc; i < Nx + 0 + Labc; i++) {
		for (int j = - Labc; j < Ny + 1 + Labc; j++) {
			memset(Hyz[i + Labc][j + Labc], 0, (Nz + 0 + 2 * Labc) * sizeof(float));
			memset(Hyx[i + Labc][j + Labc], 0, (Nz + 0 + 2 * Labc) * sizeof(float));
		}
		}
		for (int i = - Labc; i < Nx + 0 + Labc; i++) {
		for (int j = - Labc; j < Ny + 0 + Labc; j++) {
			memset(Hzx[i + Labc][j + Labc], 0, (Nz + 1 + 2 * Labc) * sizeof(float));
			memset(Hzy[i + Labc][j + Labc], 0, (Nz + 1 + 2 * Labc) * sizeof(float));
		}
		}
	}

	// debug, E
	if (iout) {
		for (int i = 0; i < Nx + 0; i++) {
		for (int j = 0; j < Ny + 1; j++) {
			memset(Ex_r[i][j], 0, (Nz + 1) * sizeof(float));
			memset(Ex_i[i][j], 0, (Nz + 1) * sizeof(float));
		}
		}
		for (int i = 0; i < Nx + 1; i++) {
		for (int j = 0; j < Ny + 0; j++) {
			memset(Ey_r[i][j], 0, (Nz + 1) * sizeof(float));
			memset(Ey_i[i][j], 0, (Nz + 1) * sizeof(float));
		}
		}
		for (int i = 0; i < Nx + 1; i++) {
		for (int j = 0; j < Ny + 1; j++) {
			memset(Ez_r[i][j], 0, (Nz + 0) * sizeof(float));
			memset(Ez_i[i][j], 0, (Nz + 0) * sizeof(float));
		}
		}
	}
}


// 平均電磁界
static void average(int Labc, int Nx, int Ny, int Nz,
	float ***Ex, float ***Ey, float ***Ez, float ***Hx, float ***Hy, float ***Hz, float fsum[])
{
	// <E>
	float sume = 0;
	int nume = 0;
	for (int i = 0; i < Nx + 0; i++) {
	for (int j = 0; j < Ny + 1; j++) {
	for (int k = 0; k < Nz + 1; k++) {
		sume += (float)fabs(Ex[i + Labc][j + Labc][k + Labc]);
		nume++;
	}
	}
	}
	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 0; j++) {
	for (int k = 0; k < Nz + 1; k++) {
		sume += (float)fabs(Ey[i + Labc][j + Labc][k + Labc]);
		nume++;
	}
	}
	}
	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 1; j++) {
	for (int k = 0; k < Nz + 0; k++) {
		sume += (float)fabs(Ez[i + Labc][j + Labc][k + Labc]);
		nume++;
	}
	}
	}
	fsum[0] = sume / nume;

	// <H>
	float sumh = 0;
	int numh = 0;
	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 0; j++) {
	for (int k = 0; k < Nz + 0; k++) {
		sumh += (float)fabs(Hx[i + Labc][j + Labc][k + Labc]);
		numh++;
	}
	}
	}
	for (int i = 0; i < Nx + 0; i++) {
	for (int j = 0; j < Ny + 1; j++) {
	for (int k = 0; k < Nz + 0; k++) {
		sumh += (float)fabs(Hy[i + Labc][j + Labc][k + Labc]);
		numh++;
	}
	}
	}
	for (int i = 0; i < Nx + 0; i++) {
	for (int j = 0; j < Ny + 0; j++) {
	for (int k = 0; k < Nz + 1; k++) {
		sumh += (float)fabs(Hz[i + Labc][j + Labc][k + Labc]);
		numh++;
	}
	}
	}
	fsum[1] = sumh / numh;
}


// DFT, debug, 全領域調和界, 計算時間が増えるので必要なときのみ
static void dft(int Labc, int Nx, int Ny, int Nz,
	float ***Ex, float ***Ey, float ***Ez, float fcos, float fsin,
	float ***Ex_r, float ***Ex_i, float ***Ey_r, float ***Ey_i, float ***Ez_r, float ***Ez_i)
{
	// Ex
	for (int i = 0; i < Nx + 0; i++) {
	for (int j = 0; j < Ny + 1; j++) {
	for (int k = 0; k < Nz + 1; k++) {
		Ex_r[i][j][k] += fcos * Ex[i + Labc][j + Labc][k + Labc];
		Ex_i[i][j][k] -= fsin * Ex[i + Labc][j + Labc][k + Labc];
	}
	}
	}

	// Ey
	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 0; j++) {
	for (int k = 0; k < Nz + 1; k++) {
		Ey_r[i][j][k] += fcos * Ey[i + Labc][j + Labc][k + Labc];
		Ey_i[i][j][k] -= fsin * Ey[i + Labc][j + Labc][k + Labc];
	}
	}
	}

	// Ez
	for (int i = 0; i < Nx + 1; i++) {
	for (int j = 0; j < Ny + 1; j++) {
	for (int k = 0; k < Nz + 0; k++) {
		Ez_r[i][j][k] += fcos * Ez[i + Labc][j + Labc][k + Labc];
		Ez_i[i][j][k] -= fsin * Ez[i + Labc][j + Labc][k + Labc];
	}
	}
	}
}
